(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('rxjs'), require('react'), require('immer')) :
  typeof define === 'function' && define.amd ? define(['exports', 'rxjs', 'react', 'immer'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.ReactRedux = {}, global.rxjs, global.React, global.produce));
})(this, (function (exports, rxjs, React, produce) { 'use strict';

  function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

  var React__default = /*#__PURE__*/_interopDefaultLegacy(React);
  var produce__default = /*#__PURE__*/_interopDefaultLegacy(produce);

  function _extends() {
    _extends = Object.assign || function (target) {
      for (var i = 1; i < arguments.length; i++) {
        var source = arguments[i];

        for (var key in source) {
          if (Object.prototype.hasOwnProperty.call(source, key)) {
            target[key] = source[key];
          }
        }
      }

      return target;
    };

    return _extends.apply(this, arguments);
  }

  class MockServer {
    constructor(config) {
      this.config = config;
    }

    getConfig() {
      const sum = this.config.responses.reduce((t, e) => t + e.ocurrence, 0);
      const ranges = this.config.responses.map(resp => {
        return _extends({}, resp, {
          percentil: Math.floor(resp.ocurrence * 100 / sum)
        });
      }).reduce((t, e) => {
        const lastRange = t.length > 0 ? t[t.length - 1] : {
          top: 0,
          bottom: 0
        };
        return [...t, _extends({}, e, {
          top: lastRange.top + e.percentil,
          bottom: lastRange.top
        })];
      }, []);
      const randomValue = Math.floor(Math.random() * 100);
      return ranges.find(range => randomValue > range.bottom && randomValue <= range.top);
    }

    invoke() {
      return rxjs.of('').pipe(rxjs.concatMap(_ => rxjs.of(this.getConfig())), rxjs.take(1), rxjs.delay(this.config.time || Math.random() * 3000), rxjs.map(config => {
        const data = config.data;
        if (config.isError) throw data;
        return data;
      }));
    }

  }

  MockServer.create = config => {
    return new MockServer(config);
  };

  exports.ProcessingType = void 0;

  (function (ProcessingType) {
    ProcessingType[ProcessingType["EXHAUST"] = 0] = "EXHAUST";
    ProcessingType[ProcessingType["CONCAT"] = 1] = "CONCAT";
    ProcessingType[ProcessingType["MERGE"] = 2] = "MERGE";
    ProcessingType[ProcessingType["SWITCH"] = 3] = "SWITCH";
    ProcessingType[ProcessingType["FLAT"] = 4] = "FLAT";
  })(exports.ProcessingType || (exports.ProcessingType = {}));

  const AsyncElement = props => {
    const status = Array.isArray(props.status) ? props.status : [props.status];
    const isProcessing = status.some(item => item.status === 'PROCESSING');
    const isCompletedOk = isProcessing ? false : status.every(item => item.status === 'SUCCESS');
    const isCompletedErrors = isProcessing ? false : status.some(item => item.status === 'ERROR');
    const elementTemplate = isCompletedOk && props.renderElement(status.map(i => i.data));
    const processingTemplate = isProcessing && props.renderProcessing && props.renderProcessing();
    const errorTemplate = isCompletedErrors && props.renderError && props.renderError(status);
    return /*#__PURE__*/React__default["default"].createElement("div", {
      className: props.className ? props.className : '',
      style: props.style ? props.style : {}
    }, elementTemplate, processingTemplate, errorTemplate);
  };

  function isFunction(value) {
      return typeof value === 'function';
  }

  function hasLift(source) {
      return isFunction(source === null || source === void 0 ? void 0 : source.lift);
  }
  function operate(init) {
      return function (source) {
          if (hasLift(source)) {
              return source.lift(function (liftedSource) {
                  try {
                      return init(liftedSource, this);
                  }
                  catch (err) {
                      this.error(err);
                  }
              });
          }
          throw new TypeError('Unable to lift unknown Observable type');
      };
  }

  /*! *****************************************************************************
  Copyright (c) Microsoft Corporation.

  Permission to use, copy, modify, and/or distribute this software for any
  purpose with or without fee is hereby granted.

  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
  REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
  AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
  INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
  PERFORMANCE OF THIS SOFTWARE.
  ***************************************************************************** */
  /* global Reflect, Promise */

  var extendStatics = function(d, b) {
      extendStatics = Object.setPrototypeOf ||
          ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
          function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
      return extendStatics(d, b);
  };

  function __extends(d, b) {
      if (typeof b !== "function" && b !== null)
          throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
      extendStatics(d, b);
      function __() { this.constructor = d; }
      d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
  }

  function __awaiter(thisArg, _arguments, P, generator) {
      function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
      return new (P || (P = Promise))(function (resolve, reject) {
          function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
          function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
          function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
          step((generator = generator.apply(thisArg, _arguments || [])).next());
      });
  }

  function __generator(thisArg, body) {
      var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
      return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
      function verb(n) { return function (v) { return step([n, v]); }; }
      function step(op) {
          if (f) throw new TypeError("Generator is already executing.");
          while (_) try {
              if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
              if (y = 0, t) op = [op[0] & 2, t.value];
              switch (op[0]) {
                  case 0: case 1: t = op; break;
                  case 4: _.label++; return { value: op[1], done: false };
                  case 5: _.label++; y = op[1]; op = [0]; continue;
                  case 7: op = _.ops.pop(); _.trys.pop(); continue;
                  default:
                      if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                      if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                      if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                      if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                      if (t[2]) _.ops.pop();
                      _.trys.pop(); continue;
              }
              op = body.call(thisArg, _);
          } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
          if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
      }
  }

  function __values(o) {
      var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
      if (m) return m.call(o);
      if (o && typeof o.length === "number") return {
          next: function () {
              if (o && i >= o.length) o = void 0;
              return { value: o && o[i++], done: !o };
          }
      };
      throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
  }

  function __read(o, n) {
      var m = typeof Symbol === "function" && o[Symbol.iterator];
      if (!m) return o;
      var i = m.call(o), r, ar = [], e;
      try {
          while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
      }
      catch (error) { e = { error: error }; }
      finally {
          try {
              if (r && !r.done && (m = i["return"])) m.call(i);
          }
          finally { if (e) throw e.error; }
      }
      return ar;
  }

  function __spreadArray(to, from) {
      for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
          to[j] = from[i];
      return to;
  }

  function __await(v) {
      return this instanceof __await ? (this.v = v, this) : new __await(v);
  }

  function __asyncGenerator(thisArg, _arguments, generator) {
      if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
      var g = generator.apply(thisArg, _arguments || []), i, q = [];
      return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
      function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
      function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
      function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }
      function fulfill(value) { resume("next", value); }
      function reject(value) { resume("throw", value); }
      function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
  }

  function __asyncValues(o) {
      if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
      var m = o[Symbol.asyncIterator], i;
      return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
      function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
      function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
  }

  var isArrayLike = (function (x) { return x && typeof x.length === 'number' && typeof x !== 'function'; });

  function isPromise(value) {
      return isFunction(value === null || value === void 0 ? void 0 : value.then);
  }

  function createErrorClass(createImpl) {
      var _super = function (instance) {
          Error.call(instance);
          instance.stack = new Error().stack;
      };
      var ctorFunc = createImpl(_super);
      ctorFunc.prototype = Object.create(Error.prototype);
      ctorFunc.prototype.constructor = ctorFunc;
      return ctorFunc;
  }

  var UnsubscriptionError = createErrorClass(function (_super) {
      return function UnsubscriptionErrorImpl(errors) {
          _super(this);
          this.message = errors
              ? errors.length + " errors occurred during unsubscription:\n" + errors.map(function (err, i) { return i + 1 + ") " + err.toString(); }).join('\n  ')
              : '';
          this.name = 'UnsubscriptionError';
          this.errors = errors;
      };
  });

  function arrRemove(arr, item) {
      if (arr) {
          var index = arr.indexOf(item);
          0 <= index && arr.splice(index, 1);
      }
  }

  var Subscription = (function () {
      function Subscription(initialTeardown) {
          this.initialTeardown = initialTeardown;
          this.closed = false;
          this._parentage = null;
          this._teardowns = null;
      }
      Subscription.prototype.unsubscribe = function () {
          var e_1, _a, e_2, _b;
          var errors;
          if (!this.closed) {
              this.closed = true;
              var _parentage = this._parentage;
              if (_parentage) {
                  this._parentage = null;
                  if (Array.isArray(_parentage)) {
                      try {
                          for (var _parentage_1 = __values(_parentage), _parentage_1_1 = _parentage_1.next(); !_parentage_1_1.done; _parentage_1_1 = _parentage_1.next()) {
                              var parent_1 = _parentage_1_1.value;
                              parent_1.remove(this);
                          }
                      }
                      catch (e_1_1) { e_1 = { error: e_1_1 }; }
                      finally {
                          try {
                              if (_parentage_1_1 && !_parentage_1_1.done && (_a = _parentage_1.return)) _a.call(_parentage_1);
                          }
                          finally { if (e_1) throw e_1.error; }
                      }
                  }
                  else {
                      _parentage.remove(this);
                  }
              }
              var initialTeardown = this.initialTeardown;
              if (isFunction(initialTeardown)) {
                  try {
                      initialTeardown();
                  }
                  catch (e) {
                      errors = e instanceof UnsubscriptionError ? e.errors : [e];
                  }
              }
              var _teardowns = this._teardowns;
              if (_teardowns) {
                  this._teardowns = null;
                  try {
                      for (var _teardowns_1 = __values(_teardowns), _teardowns_1_1 = _teardowns_1.next(); !_teardowns_1_1.done; _teardowns_1_1 = _teardowns_1.next()) {
                          var teardown_1 = _teardowns_1_1.value;
                          try {
                              execTeardown(teardown_1);
                          }
                          catch (err) {
                              errors = errors !== null && errors !== void 0 ? errors : [];
                              if (err instanceof UnsubscriptionError) {
                                  errors = __spreadArray(__spreadArray([], __read(errors), false), __read(err.errors), false);
                              }
                              else {
                                  errors.push(err);
                              }
                          }
                      }
                  }
                  catch (e_2_1) { e_2 = { error: e_2_1 }; }
                  finally {
                      try {
                          if (_teardowns_1_1 && !_teardowns_1_1.done && (_b = _teardowns_1.return)) _b.call(_teardowns_1);
                      }
                      finally { if (e_2) throw e_2.error; }
                  }
              }
              if (errors) {
                  throw new UnsubscriptionError(errors);
              }
          }
      };
      Subscription.prototype.add = function (teardown) {
          var _a;
          if (teardown && teardown !== this) {
              if (this.closed) {
                  execTeardown(teardown);
              }
              else {
                  if (teardown instanceof Subscription) {
                      if (teardown.closed || teardown._hasParent(this)) {
                          return;
                      }
                      teardown._addParent(this);
                  }
                  (this._teardowns = (_a = this._teardowns) !== null && _a !== void 0 ? _a : []).push(teardown);
              }
          }
      };
      Subscription.prototype._hasParent = function (parent) {
          var _parentage = this._parentage;
          return _parentage === parent || (Array.isArray(_parentage) && _parentage.includes(parent));
      };
      Subscription.prototype._addParent = function (parent) {
          var _parentage = this._parentage;
          this._parentage = Array.isArray(_parentage) ? (_parentage.push(parent), _parentage) : _parentage ? [_parentage, parent] : parent;
      };
      Subscription.prototype._removeParent = function (parent) {
          var _parentage = this._parentage;
          if (_parentage === parent) {
              this._parentage = null;
          }
          else if (Array.isArray(_parentage)) {
              arrRemove(_parentage, parent);
          }
      };
      Subscription.prototype.remove = function (teardown) {
          var _teardowns = this._teardowns;
          _teardowns && arrRemove(_teardowns, teardown);
          if (teardown instanceof Subscription) {
              teardown._removeParent(this);
          }
      };
      Subscription.EMPTY = (function () {
          var empty = new Subscription();
          empty.closed = true;
          return empty;
      })();
      return Subscription;
  }());
  function isSubscription(value) {
      return (value instanceof Subscription ||
          (value && 'closed' in value && isFunction(value.remove) && isFunction(value.add) && isFunction(value.unsubscribe)));
  }
  function execTeardown(teardown) {
      if (isFunction(teardown)) {
          teardown();
      }
      else {
          teardown.unsubscribe();
      }
  }

  var config = {
      onUnhandledError: null,
      onStoppedNotification: null,
      Promise: undefined,
      useDeprecatedSynchronousErrorHandling: false,
      useDeprecatedNextContext: false,
  };

  var timeoutProvider = {
      setTimeout: function () {
          var args = [];
          for (var _i = 0; _i < arguments.length; _i++) {
              args[_i] = arguments[_i];
          }
          var delegate = timeoutProvider.delegate;
          return ((delegate === null || delegate === void 0 ? void 0 : delegate.setTimeout) || setTimeout).apply(void 0, __spreadArray([], __read(args)));
      },
      clearTimeout: function (handle) {
          var delegate = timeoutProvider.delegate;
          return ((delegate === null || delegate === void 0 ? void 0 : delegate.clearTimeout) || clearTimeout)(handle);
      },
      delegate: undefined,
  };

  function reportUnhandledError(err) {
      timeoutProvider.setTimeout(function () {
          var onUnhandledError = config.onUnhandledError;
          if (onUnhandledError) {
              onUnhandledError(err);
          }
          else {
              throw err;
          }
      });
  }

  function noop() { }

  var COMPLETE_NOTIFICATION = (function () { return createNotification('C', undefined, undefined); })();
  function errorNotification(error) {
      return createNotification('E', undefined, error);
  }
  function nextNotification(value) {
      return createNotification('N', value, undefined);
  }
  function createNotification(kind, value, error) {
      return {
          kind: kind,
          value: value,
          error: error,
      };
  }

  var context = null;
  function errorContext(cb) {
      {
          cb();
      }
  }
  function captureError(err) {
      if (config.useDeprecatedSynchronousErrorHandling && context) {
          context.errorThrown = true;
          context.error = err;
      }
  }

  var Subscriber = (function (_super) {
      __extends(Subscriber, _super);
      function Subscriber(destination) {
          var _this = _super.call(this) || this;
          _this.isStopped = false;
          if (destination) {
              _this.destination = destination;
              if (isSubscription(destination)) {
                  destination.add(_this);
              }
          }
          else {
              _this.destination = EMPTY_OBSERVER;
          }
          return _this;
      }
      Subscriber.create = function (next, error, complete) {
          return new SafeSubscriber(next, error, complete);
      };
      Subscriber.prototype.next = function (value) {
          if (this.isStopped) {
              handleStoppedNotification(nextNotification(value), this);
          }
          else {
              this._next(value);
          }
      };
      Subscriber.prototype.error = function (err) {
          if (this.isStopped) {
              handleStoppedNotification(errorNotification(err), this);
          }
          else {
              this.isStopped = true;
              this._error(err);
          }
      };
      Subscriber.prototype.complete = function () {
          if (this.isStopped) {
              handleStoppedNotification(COMPLETE_NOTIFICATION, this);
          }
          else {
              this.isStopped = true;
              this._complete();
          }
      };
      Subscriber.prototype.unsubscribe = function () {
          if (!this.closed) {
              this.isStopped = true;
              _super.prototype.unsubscribe.call(this);
              this.destination = null;
          }
      };
      Subscriber.prototype._next = function (value) {
          this.destination.next(value);
      };
      Subscriber.prototype._error = function (err) {
          try {
              this.destination.error(err);
          }
          finally {
              this.unsubscribe();
          }
      };
      Subscriber.prototype._complete = function () {
          try {
              this.destination.complete();
          }
          finally {
              this.unsubscribe();
          }
      };
      return Subscriber;
  }(Subscription));
  var SafeSubscriber = (function (_super) {
      __extends(SafeSubscriber, _super);
      function SafeSubscriber(observerOrNext, error, complete) {
          var _this = _super.call(this) || this;
          var next;
          if (isFunction(observerOrNext)) {
              next = observerOrNext;
          }
          else if (observerOrNext) {
              (next = observerOrNext.next, error = observerOrNext.error, complete = observerOrNext.complete);
              var context_1;
              if (_this && config.useDeprecatedNextContext) {
                  context_1 = Object.create(observerOrNext);
                  context_1.unsubscribe = function () { return _this.unsubscribe(); };
              }
              else {
                  context_1 = observerOrNext;
              }
              next = next === null || next === void 0 ? void 0 : next.bind(context_1);
              error = error === null || error === void 0 ? void 0 : error.bind(context_1);
              complete = complete === null || complete === void 0 ? void 0 : complete.bind(context_1);
          }
          _this.destination = {
              next: next ? wrapForErrorHandling(next) : noop,
              error: wrapForErrorHandling(error !== null && error !== void 0 ? error : defaultErrorHandler),
              complete: complete ? wrapForErrorHandling(complete) : noop,
          };
          return _this;
      }
      return SafeSubscriber;
  }(Subscriber));
  function wrapForErrorHandling(handler, instance) {
      return function () {
          var args = [];
          for (var _i = 0; _i < arguments.length; _i++) {
              args[_i] = arguments[_i];
          }
          try {
              handler.apply(void 0, __spreadArray([], __read(args), false));
          }
          catch (err) {
              if (config.useDeprecatedSynchronousErrorHandling) {
                  captureError(err);
              }
              else {
                  reportUnhandledError(err);
              }
          }
      };
  }
  function defaultErrorHandler(err) {
      throw err;
  }
  function handleStoppedNotification(notification, subscriber) {
      var onStoppedNotification = config.onStoppedNotification;
      onStoppedNotification && timeoutProvider.setTimeout(function () { return onStoppedNotification(notification, subscriber); });
  }
  var EMPTY_OBSERVER = {
      closed: true,
      next: noop,
      error: defaultErrorHandler,
      complete: noop,
  };

  var observable = (function () { return (typeof Symbol === 'function' && Symbol.observable) || '@@observable'; })();

  function identity(x) {
      return x;
  }

  function pipeFromArray(fns) {
      if (fns.length === 0) {
          return identity;
      }
      if (fns.length === 1) {
          return fns[0];
      }
      return function piped(input) {
          return fns.reduce(function (prev, fn) { return fn(prev); }, input);
      };
  }

  var Observable = (function () {
      function Observable(subscribe) {
          if (subscribe) {
              this._subscribe = subscribe;
          }
      }
      Observable.prototype.lift = function (operator) {
          var observable = new Observable();
          observable.source = this;
          observable.operator = operator;
          return observable;
      };
      Observable.prototype.subscribe = function (observerOrNext, error, complete) {
          var _this = this;
          var subscriber = isSubscriber(observerOrNext) ? observerOrNext : new SafeSubscriber(observerOrNext, error, complete);
          errorContext(function () {
              var _a = _this, operator = _a.operator, source = _a.source;
              subscriber.add(operator
                  ?
                      operator.call(subscriber, source)
                  : source
                      ?
                          _this._subscribe(subscriber)
                      :
                          _this._trySubscribe(subscriber));
          });
          return subscriber;
      };
      Observable.prototype._trySubscribe = function (sink) {
          try {
              return this._subscribe(sink);
          }
          catch (err) {
              sink.error(err);
          }
      };
      Observable.prototype.forEach = function (next, promiseCtor) {
          var _this = this;
          promiseCtor = getPromiseCtor(promiseCtor);
          return new promiseCtor(function (resolve, reject) {
              var subscription;
              subscription = _this.subscribe(function (value) {
                  try {
                      next(value);
                  }
                  catch (err) {
                      reject(err);
                      subscription === null || subscription === void 0 ? void 0 : subscription.unsubscribe();
                  }
              }, reject, resolve);
          });
      };
      Observable.prototype._subscribe = function (subscriber) {
          var _a;
          return (_a = this.source) === null || _a === void 0 ? void 0 : _a.subscribe(subscriber);
      };
      Observable.prototype[observable] = function () {
          return this;
      };
      Observable.prototype.pipe = function () {
          var operations = [];
          for (var _i = 0; _i < arguments.length; _i++) {
              operations[_i] = arguments[_i];
          }
          return pipeFromArray(operations)(this);
      };
      Observable.prototype.toPromise = function (promiseCtor) {
          var _this = this;
          promiseCtor = getPromiseCtor(promiseCtor);
          return new promiseCtor(function (resolve, reject) {
              var value;
              _this.subscribe(function (x) { return (value = x); }, function (err) { return reject(err); }, function () { return resolve(value); });
          });
      };
      Observable.create = function (subscribe) {
          return new Observable(subscribe);
      };
      return Observable;
  }());
  function getPromiseCtor(promiseCtor) {
      var _a;
      return (_a = promiseCtor !== null && promiseCtor !== void 0 ? promiseCtor : config.Promise) !== null && _a !== void 0 ? _a : Promise;
  }
  function isObserver(value) {
      return value && isFunction(value.next) && isFunction(value.error) && isFunction(value.complete);
  }
  function isSubscriber(value) {
      return (value && value instanceof Subscriber) || (isObserver(value) && isSubscription(value));
  }

  function isInteropObservable(input) {
      return isFunction(input[observable]);
  }

  function isAsyncIterable(obj) {
      return Symbol.asyncIterator && isFunction(obj === null || obj === void 0 ? void 0 : obj[Symbol.asyncIterator]);
  }

  function createInvalidObservableTypeError(input) {
      return new TypeError("You provided " + (input !== null && typeof input === 'object' ? 'an invalid object' : "'" + input + "'") + " where a stream was expected. You can provide an Observable, Promise, ReadableStream, Array, AsyncIterable, or Iterable.");
  }

  function getSymbolIterator() {
      if (typeof Symbol !== 'function' || !Symbol.iterator) {
          return '@@iterator';
      }
      return Symbol.iterator;
  }
  var iterator = getSymbolIterator();

  function isIterable(input) {
      return isFunction(input === null || input === void 0 ? void 0 : input[iterator]);
  }

  function readableStreamLikeToAsyncGenerator(readableStream) {
      return __asyncGenerator(this, arguments, function readableStreamLikeToAsyncGenerator_1() {
          var reader, _a, value, done;
          return __generator(this, function (_b) {
              switch (_b.label) {
                  case 0:
                      reader = readableStream.getReader();
                      _b.label = 1;
                  case 1:
                      _b.trys.push([1, , 9, 10]);
                      _b.label = 2;
                  case 2:
                      return [4, __await(reader.read())];
                  case 3:
                      _a = _b.sent(), value = _a.value, done = _a.done;
                      if (!done) return [3, 5];
                      return [4, __await(void 0)];
                  case 4: return [2, _b.sent()];
                  case 5: return [4, __await(value)];
                  case 6: return [4, _b.sent()];
                  case 7:
                      _b.sent();
                      return [3, 2];
                  case 8: return [3, 10];
                  case 9:
                      reader.releaseLock();
                      return [7];
                  case 10: return [2];
              }
          });
      });
  }
  function isReadableStreamLike(obj) {
      return isFunction(obj === null || obj === void 0 ? void 0 : obj.getReader);
  }

  function innerFrom(input) {
      if (input instanceof Observable) {
          return input;
      }
      if (input != null) {
          if (isInteropObservable(input)) {
              return fromInteropObservable(input);
          }
          if (isArrayLike(input)) {
              return fromArrayLike(input);
          }
          if (isPromise(input)) {
              return fromPromise(input);
          }
          if (isAsyncIterable(input)) {
              return fromAsyncIterable(input);
          }
          if (isIterable(input)) {
              return fromIterable(input);
          }
          if (isReadableStreamLike(input)) {
              return fromReadableStreamLike(input);
          }
      }
      throw createInvalidObservableTypeError(input);
  }
  function fromInteropObservable(obj) {
      return new Observable(function (subscriber) {
          var obs = obj[observable]();
          if (isFunction(obs.subscribe)) {
              return obs.subscribe(subscriber);
          }
          throw new TypeError('Provided object does not correctly implement Symbol.observable');
      });
  }
  function fromArrayLike(array) {
      return new Observable(function (subscriber) {
          for (var i = 0; i < array.length && !subscriber.closed; i++) {
              subscriber.next(array[i]);
          }
          subscriber.complete();
      });
  }
  function fromPromise(promise) {
      return new Observable(function (subscriber) {
          promise
              .then(function (value) {
              if (!subscriber.closed) {
                  subscriber.next(value);
                  subscriber.complete();
              }
          }, function (err) { return subscriber.error(err); })
              .then(null, reportUnhandledError);
      });
  }
  function fromIterable(iterable) {
      return new Observable(function (subscriber) {
          var e_1, _a;
          try {
              for (var iterable_1 = __values(iterable), iterable_1_1 = iterable_1.next(); !iterable_1_1.done; iterable_1_1 = iterable_1.next()) {
                  var value = iterable_1_1.value;
                  subscriber.next(value);
                  if (subscriber.closed) {
                      return;
                  }
              }
          }
          catch (e_1_1) { e_1 = { error: e_1_1 }; }
          finally {
              try {
                  if (iterable_1_1 && !iterable_1_1.done && (_a = iterable_1.return)) _a.call(iterable_1);
              }
              finally { if (e_1) throw e_1.error; }
          }
          subscriber.complete();
      });
  }
  function fromAsyncIterable(asyncIterable) {
      return new Observable(function (subscriber) {
          process(asyncIterable, subscriber).catch(function (err) { return subscriber.error(err); });
      });
  }
  function fromReadableStreamLike(readableStream) {
      return fromAsyncIterable(readableStreamLikeToAsyncGenerator(readableStream));
  }
  function process(asyncIterable, subscriber) {
      var asyncIterable_1, asyncIterable_1_1;
      var e_2, _a;
      return __awaiter(this, void 0, void 0, function () {
          var value, e_2_1;
          return __generator(this, function (_b) {
              switch (_b.label) {
                  case 0:
                      _b.trys.push([0, 5, 6, 11]);
                      asyncIterable_1 = __asyncValues(asyncIterable);
                      _b.label = 1;
                  case 1: return [4, asyncIterable_1.next()];
                  case 2:
                      if (!(asyncIterable_1_1 = _b.sent(), !asyncIterable_1_1.done)) return [3, 4];
                      value = asyncIterable_1_1.value;
                      subscriber.next(value);
                      if (subscriber.closed) {
                          return [2];
                      }
                      _b.label = 3;
                  case 3: return [3, 1];
                  case 4: return [3, 11];
                  case 5:
                      e_2_1 = _b.sent();
                      e_2 = { error: e_2_1 };
                      return [3, 11];
                  case 6:
                      _b.trys.push([6, , 9, 10]);
                      if (!(asyncIterable_1_1 && !asyncIterable_1_1.done && (_a = asyncIterable_1.return))) return [3, 8];
                      return [4, _a.call(asyncIterable_1)];
                  case 7:
                      _b.sent();
                      _b.label = 8;
                  case 8: return [3, 10];
                  case 9:
                      if (e_2) throw e_2.error;
                      return [7];
                  case 10: return [7];
                  case 11:
                      subscriber.complete();
                      return [2];
              }
          });
      });
  }

  var OperatorSubscriber = (function (_super) {
      __extends(OperatorSubscriber, _super);
      function OperatorSubscriber(destination, onNext, onComplete, onError, onFinalize) {
          var _this = _super.call(this, destination) || this;
          _this.onFinalize = onFinalize;
          _this._next = onNext
              ? function (value) {
                  try {
                      onNext(value);
                  }
                  catch (err) {
                      destination.error(err);
                  }
              }
              : _super.prototype._next;
          _this._error = onError
              ? function (err) {
                  try {
                      onError(err);
                  }
                  catch (err) {
                      destination.error(err);
                  }
                  finally {
                      this.unsubscribe();
                  }
              }
              : _super.prototype._error;
          _this._complete = onComplete
              ? function () {
                  try {
                      onComplete();
                  }
                  catch (err) {
                      destination.error(err);
                  }
                  finally {
                      this.unsubscribe();
                  }
              }
              : _super.prototype._complete;
          return _this;
      }
      OperatorSubscriber.prototype.unsubscribe = function () {
          var _a;
          var closed = this.closed;
          _super.prototype.unsubscribe.call(this);
          !closed && ((_a = this.onFinalize) === null || _a === void 0 ? void 0 : _a.call(this));
      };
      return OperatorSubscriber;
  }(Subscriber));

  function isScheduler(value) {
      return value && isFunction(value.schedule);
  }

  function last(arr) {
      return arr[arr.length - 1];
  }
  function popScheduler(args) {
      return isScheduler(last(args)) ? args.pop() : undefined;
  }

  function executeSchedule(parentSubscription, scheduler, work, delay, repeat) {
      if (delay === void 0) { delay = 0; }
      if (repeat === void 0) { repeat = false; }
      var scheduleSubscription = scheduler.schedule(function () {
          work();
          if (repeat) {
              parentSubscription.add(this.schedule(null, delay));
          }
          else {
              this.unsubscribe();
          }
      }, delay);
      parentSubscription.add(scheduleSubscription);
      if (!repeat) {
          return scheduleSubscription;
      }
  }

  function observeOn(scheduler, delay) {
      if (delay === void 0) { delay = 0; }
      return operate(function (source, subscriber) {
          source.subscribe(new OperatorSubscriber(subscriber, function (value) { return executeSchedule(subscriber, scheduler, function () { return subscriber.next(value); }, delay); }, function () { return executeSchedule(subscriber, scheduler, function () { return subscriber.complete(); }, delay); }, function (err) { return executeSchedule(subscriber, scheduler, function () { return subscriber.error(err); }, delay); }));
      });
  }

  function subscribeOn(scheduler, delay) {
      if (delay === void 0) { delay = 0; }
      return operate(function (source, subscriber) {
          subscriber.add(scheduler.schedule(function () { return source.subscribe(subscriber); }, delay));
      });
  }

  function scheduleObservable(input, scheduler) {
      return innerFrom(input).pipe(subscribeOn(scheduler), observeOn(scheduler));
  }

  function schedulePromise(input, scheduler) {
      return innerFrom(input).pipe(subscribeOn(scheduler), observeOn(scheduler));
  }

  function scheduleArray(input, scheduler) {
      return new Observable(function (subscriber) {
          var i = 0;
          return scheduler.schedule(function () {
              if (i === input.length) {
                  subscriber.complete();
              }
              else {
                  subscriber.next(input[i++]);
                  if (!subscriber.closed) {
                      this.schedule();
                  }
              }
          });
      });
  }

  function scheduleIterable(input, scheduler) {
      return new Observable(function (subscriber) {
          var iterator$1;
          executeSchedule(subscriber, scheduler, function () {
              iterator$1 = input[iterator]();
              executeSchedule(subscriber, scheduler, function () {
                  var _a;
                  var value;
                  var done;
                  try {
                      (_a = iterator$1.next(), value = _a.value, done = _a.done);
                  }
                  catch (err) {
                      subscriber.error(err);
                      return;
                  }
                  if (done) {
                      subscriber.complete();
                  }
                  else {
                      subscriber.next(value);
                  }
              }, 0, true);
          });
          return function () { return isFunction(iterator$1 === null || iterator$1 === void 0 ? void 0 : iterator$1.return) && iterator$1.return(); };
      });
  }

  function scheduleAsyncIterable(input, scheduler) {
      if (!input) {
          throw new Error('Iterable cannot be null');
      }
      return new Observable(function (subscriber) {
          executeSchedule(subscriber, scheduler, function () {
              var iterator = input[Symbol.asyncIterator]();
              executeSchedule(subscriber, scheduler, function () {
                  iterator.next().then(function (result) {
                      if (result.done) {
                          subscriber.complete();
                      }
                      else {
                          subscriber.next(result.value);
                      }
                  });
              }, 0, true);
          });
      });
  }

  function scheduleReadableStreamLike(input, scheduler) {
      return scheduleAsyncIterable(readableStreamLikeToAsyncGenerator(input), scheduler);
  }

  function scheduled(input, scheduler) {
      if (input != null) {
          if (isInteropObservable(input)) {
              return scheduleObservable(input, scheduler);
          }
          if (isArrayLike(input)) {
              return scheduleArray(input, scheduler);
          }
          if (isPromise(input)) {
              return schedulePromise(input, scheduler);
          }
          if (isAsyncIterable(input)) {
              return scheduleAsyncIterable(input, scheduler);
          }
          if (isIterable(input)) {
              return scheduleIterable(input, scheduler);
          }
          if (isReadableStreamLike(input)) {
              return scheduleReadableStreamLike(input, scheduler);
          }
      }
      throw createInvalidObservableTypeError(input);
  }

  function from(input, scheduler) {
      return scheduler ? scheduled(input, scheduler) : innerFrom(input);
  }

  function map(project, thisArg) {
      return operate(function (source, subscriber) {
          var index = 0;
          source.subscribe(new OperatorSubscriber(subscriber, function (value) {
              subscriber.next(project.call(thisArg, value, index++));
          }));
      });
  }

  function mergeInternals(source, subscriber, project, concurrent, onBeforeNext, expand, innerSubScheduler, additionalTeardown) {
      var buffer = [];
      var active = 0;
      var index = 0;
      var isComplete = false;
      var checkComplete = function () {
          if (isComplete && !buffer.length && !active) {
              subscriber.complete();
          }
      };
      var outerNext = function (value) { return (active < concurrent ? doInnerSub(value) : buffer.push(value)); };
      var doInnerSub = function (value) {
          expand && subscriber.next(value);
          active++;
          var innerComplete = false;
          innerFrom(project(value, index++)).subscribe(new OperatorSubscriber(subscriber, function (innerValue) {
              onBeforeNext === null || onBeforeNext === void 0 ? void 0 : onBeforeNext(innerValue);
              if (expand) {
                  outerNext(innerValue);
              }
              else {
                  subscriber.next(innerValue);
              }
          }, function () {
              innerComplete = true;
          }, undefined, function () {
              if (innerComplete) {
                  try {
                      active--;
                      var _loop_1 = function () {
                          var bufferedValue = buffer.shift();
                          if (innerSubScheduler) {
                              executeSchedule(subscriber, innerSubScheduler, function () { return doInnerSub(bufferedValue); });
                          }
                          else {
                              doInnerSub(bufferedValue);
                          }
                      };
                      while (buffer.length && active < concurrent) {
                          _loop_1();
                      }
                      checkComplete();
                  }
                  catch (err) {
                      subscriber.error(err);
                  }
              }
          }));
      };
      source.subscribe(new OperatorSubscriber(subscriber, outerNext, function () {
          isComplete = true;
          checkComplete();
      }));
      return function () {
          additionalTeardown === null || additionalTeardown === void 0 ? void 0 : additionalTeardown();
      };
  }

  function mergeMap(project, resultSelector, concurrent) {
      if (concurrent === void 0) { concurrent = Infinity; }
      if (isFunction(resultSelector)) {
          return mergeMap(function (a, i) { return map(function (b, ii) { return resultSelector(a, b, i, ii); })(innerFrom(project(a, i))); }, concurrent);
      }
      else if (typeof resultSelector === 'number') {
          concurrent = resultSelector;
      }
      return operate(function (source, subscriber) { return mergeInternals(source, subscriber, project, concurrent); });
  }

  function mergeAll(concurrent) {
      if (concurrent === void 0) { concurrent = Infinity; }
      return mergeMap(identity, concurrent);
  }

  function concatAll() {
      return mergeAll(1);
  }

  function concat() {
      var args = [];
      for (var _i = 0; _i < arguments.length; _i++) {
          args[_i] = arguments[_i];
      }
      return concatAll()(from(args, popScheduler(args)));
  }

  function startWith() {
      var values = [];
      for (var _i = 0; _i < arguments.length; _i++) {
          values[_i] = arguments[_i];
      }
      var scheduler = popScheduler(values);
      return operate(function (source, subscriber) {
          (scheduler ? concat(values, source, scheduler) : concat(values, source)).subscribe(subscriber);
      });
  }

  function takeUntil(notifier) {
      return operate(function (source, subscriber) {
          innerFrom(notifier).subscribe(new OperatorSubscriber(subscriber, function () { return subscriber.complete(); }, noop));
          !subscriber.closed && source.subscribe(subscriber);
      });
  }

  function tap(observerOrNext, error, complete) {
      var tapObserver = isFunction(observerOrNext) || error || complete
          ?
              { next: observerOrNext, error: error, complete: complete }
          : observerOrNext;
      return tapObserver
          ? operate(function (source, subscriber) {
              var _a;
              (_a = tapObserver.subscribe) === null || _a === void 0 ? void 0 : _a.call(tapObserver);
              var isUnsub = true;
              source.subscribe(new OperatorSubscriber(subscriber, function (value) {
                  var _a;
                  (_a = tapObserver.next) === null || _a === void 0 ? void 0 : _a.call(tapObserver, value);
                  subscriber.next(value);
              }, function () {
                  var _a;
                  isUnsub = false;
                  (_a = tapObserver.complete) === null || _a === void 0 ? void 0 : _a.call(tapObserver);
                  subscriber.complete();
              }, function (err) {
                  var _a;
                  isUnsub = false;
                  (_a = tapObserver.error) === null || _a === void 0 ? void 0 : _a.call(tapObserver, err);
                  subscriber.error(err);
              }, function () {
                  var _a, _b;
                  if (isUnsub) {
                      (_a = tapObserver.unsubscribe) === null || _a === void 0 ? void 0 : _a.call(tapObserver);
                  }
                  (_b = tapObserver.finalize) === null || _b === void 0 ? void 0 : _b.call(tapObserver);
              }));
          })
          :
              identity;
  }

  class InputContainer extends React.PureComponent {
    constructor(props) {
      super(props);
      this.destroy$ = new rxjs.Subject();
      this.state = {
        status: 'SUCCESS'
      };
      this.initializeControl.bind(this);
    }

    componentDidMount() {
      const control = this.props.control;
      control && this.initializeControl(control);
    }

    initializeControl(control) {
      control.statusChanges.pipe(takeUntil(this.destroy$), startWith(control.status), map(statusType => {
        if (statusType === 'INVALID') {
          return {
            status: 'ERROR',
            error: control.errors
          };
        } else if (statusType === 'PENDING') {
          return {
            status: 'PROCESSING',
            data: {}
          };
        }

        return {
          status: 'SUCCESS',
          data: {}
        };
      }), tap(status => {
        this.setState(produce__default["default"](draft => {
          draft.data = status.data;
          draft.error = status.error;
          draft.status = status.status;
        }));
      })).subscribe();
    }

    componentDidUpdate(prevProps, prevState) {
      let newControl = this.props.control;

      if (newControl && prevProps.control !== newControl) {
        this.initializeControl(newControl);
      }
    }

    componentWillUnmount() {
      this.destroy$.next();
      this.destroy$.complete();
    }

    render() {
      const control = this.props.control;
      const showErrorsFn = this.props.showErrorsFn ? this.props.showErrorsFn : () => true;
      const formLoadStatus = this.props.formLoadStatus ? this.props.formLoadStatus : 'SUCCESS';
      const showErrors = control ? showErrorsFn(control) : false;
      const data = this.props.data ? this.props.data : {};
      const dataLoadStatus = data.load ? data.load : {
        status: 'SUCCESS'
      };
      const renderLabel = this.props.renderLabel;
      const labelTemplate = renderLabel && renderLabel({
        showErrors,
        validationStatus: this.state.status
      });
      const renderInput = this.props.renderInput;
      const inputTemplate = control && formLoadStatus === 'SUCCESS' && dataLoadStatus.status === 'SUCCESS' && renderInput({
        control,
        data: dataLoadStatus.data,
        validationStatus: this.state.status,
        showErrors
      });
      const renderProcessing = this.props.renderProcessing; // if(formLoadStatus==='PROCESSING'){
      //     debugger;
      // }

      const processingTemplate = (formLoadStatus === 'PROCESSING' || dataLoadStatus.status === 'PROCESSING') && renderProcessing && renderProcessing();
      const renderValidationErrors = this.props.renderValidationErrors;
      const validationErrorsTemplate = control && formLoadStatus === 'SUCCESS' && dataLoadStatus.status === 'SUCCESS' && showErrors && renderValidationErrors && renderValidationErrors({
        errors: this.state.error,
        control
      });
      const renderReload = data.renderReload;
      const reloadTemplate = dataLoadStatus.status === 'ERROR' && renderReload && renderReload(dataLoadStatus.error);
      return /*#__PURE__*/React__default["default"].createElement(React__default["default"].Fragment, null, formLoadStatus !== 'ERROR' && /*#__PURE__*/React__default["default"].createElement(React__default["default"].Fragment, null, labelTemplate, inputTemplate, processingTemplate, validationErrorsTemplate, reloadTemplate), formLoadStatus === 'SUCCESS' && dataLoadStatus.status === 'SUCCESS' && !this.props.control && /*#__PURE__*/React__default["default"].createElement(React__default["default"].Fragment, null));
    }

  }

  let ModuleUtils;

  (function (_ModuleUtils) {
    function isPromise(obj) {
      return !!obj && typeof obj.then === 'function';
    }

    _ModuleUtils.isPromise = isPromise;

    function isSubscribable(obj) {
      return !!obj && typeof obj.subscribe === 'function';
    }

    _ModuleUtils.isSubscribable = isSubscribable;
    _ModuleUtils.isObservable = isSubscribable;
  })(ModuleUtils || (ModuleUtils = {}));

  const initialStateAsyncCall = {
    params: {},
    status: {},
    errors: []
  };
  const getTypeOperation = processingType => {
    switch (processingType) {
      case exports.ProcessingType.CONCAT:
        return rxjs.concatMap;

      case exports.ProcessingType.EXHAUST:
        return rxjs.exhaustMap;

      case exports.ProcessingType.FLAT:
        return rxjs.flatMap;

      case exports.ProcessingType.MERGE:
        return rxjs.mergeMap;

      case exports.ProcessingType.SWITCH:
        return rxjs.switchMap;

      default:
        return rxjs.switchMap;
    }
  };
  class AsyncCall {
    constructor(executeFn$, config, setState, destroy$) {
      this.executeFn$ = executeFn$;
      this.config = config;
      this.setState = setState;
      this.destroy$ = destroy$;
      this.execute$ = new rxjs.Subject();
      this.newState = initialStateAsyncCall;
      this.setData$Subject = new rxjs.Subject();
      this.destroy$ = destroy$ ? destroy$ : new rxjs.Subject();
      this.init = this.init.bind(this);
      this.getState = this.getState.bind(this);
      this.reload = this.reload.bind(this);
      this.initSetData = this.initSetData.bind(this);
      this.getStatus = this.getStatus.bind(this);
    }

    getStatus() {
      return this.newState.status.status;
    }

    initSetData() {
      this.setData$Subject.pipe(rxjs.takeUntil(this.destroy$), rxjs.tap(setData => {
        this.setState(prevState => {
          const newState = produce__default["default"](draft => {
            const state = this.getState();
            debugger;
            draft.status.data = setData(state.status.data);
          })(prevState);
          this.newState = newState;
          return newState;
        });
      })).subscribe();
    }

    init() {
      this.initSetData();
      const processingStatus$ = new rxjs.Subject();
      const config = this.config ? this.config : {};
      const params$ = config.source$ ? rxjs.merge(this.config.source$, this.execute$) : this.execute$;
      const asyncOperation = getTypeOperation(config.processingType);
      rxjs.merge(processingStatus$.pipe(rxjs.mapTo({
        status: 'PROCESSING'
      })), params$.pipe(rxjs.takeUntil(this.destroy$), rxjs.debounceTime(10), rxjs.tap(params => {
        this.setState(prevState => {
          const newState = produce__default["default"](draft => {
            draft.params = params;
            draft.errors = [];
          })(prevState);
          this.newState = newState;
          return newState;
        });
        processingStatus$.next();
      }), asyncOperation(params => {
        const data$ = this.executeFn$ ? this.executeFn$(params) : rxjs.of(params);
        const loadData$ = ModuleUtils.isObservable(data$) ? data$ : ModuleUtils.isPromise(data$) ? rxjs.from(data$) : rxjs.of(data$);
        const retryConfig = config.retry ? config.retry : {
          interval: 1000,
          maxRetryAttempts: 0
        };
        const maxRetryAttempts = retryConfig.maxRetryAttempts ? retryConfig.maxRetryAttempts : 0;
        const retryWhenStatus = retryConfig.retryWhenStatus ? retryConfig.retryWhenStatus : [];
        const notRetryWhenStatus = retryConfig.notRetryWhenStatus ? retryConfig.notRetryWhenStatus : [];
        return loadData$.pipe(rxjs.retryWhen(errors => errors.pipe(rxjs.tap(e => {
          this.setState(prevState => {
            const newState = produce__default["default"](draft => {
              draft.errors.push(e);
            })(prevState);
            this.newState = newState;
            return newState;
          });
        }), rxjs.concatMap((e, i) => rxjs.iif(() => {
          if (i < maxRetryAttempts) {
            const responseStatus = (e || {}).status || 0;
            const retry = !!retryWhenStatus.find(status => status === responseStatus);

            if (retry) {
              return false;
            }

            const notRetry = !!notRetryWhenStatus.find(status => status === responseStatus);

            if (notRetry) {
              return true;
            }

            if (retryConfig.retryWhen && retryConfig.retryWhen(e)) {
              return false;
            }

            if (retryConfig.noRetryWhen && retryConfig.noRetryWhen(e)) {
              return true;
            }

            return false;
          }

          return true;
        }, rxjs.throwError(e), rxjs.of(e).pipe(rxjs.delay((retryConfig.typeInterval === 'LINEAR' && 1 || retryConfig.typeInterval === 'EXPONENTIAL' && i) * retryConfig.interval)))))), rxjs.take(1), rxjs.map(data => {
          return {
            status: 'SUCCESS',
            data
          };
        }), rxjs.catchError(error => {
          return rxjs.of({
            status: 'ERROR',
            error
          });
        }));
      }))).pipe(rxjs.takeUntil(this.destroy$), rxjs.tap(processingStatus => {
        this.setState(prevState => {
          const newState = produce__default["default"](draft => {
            draft.status.data = processingStatus.data;
            draft.status.error = processingStatus.error;
          })(prevState);
          this.newState = newState;
          return newState;
        });
      }), rxjs.tap(processingStatus => {
        const params = this.newState.params;
        processingStatus.status === 'PROCESSING' && config.onProcessing && config.onProcessing(params);
        processingStatus.status === 'ERROR' && config.onError && config.onError(processingStatus.error, params);
        processingStatus.status === 'SUCCESS' && config.onSuccess && config.onSuccess(processingStatus.data, params);
        this.setState(prevState => {
          const newState = produce__default["default"](draft => {
            draft.status.status = processingStatus.status;
          })(prevState);
          this.newState = newState;
          return newState;
        });
      })).subscribe();
    }

    getState() {
      return this.newState;
    }

    reload() {
      this.execute$.next(this.newState.params);
    }

    get execute() {
      return this.execute$.next.bind(this.execute$);
    }

    get setData() {
      return this.setData$Subject.next.bind(this.setData$Subject);
    }

  }

  const useAsyncCall = (executeFn, config) => {
    const [, setState] = React.useState(initialStateAsyncCall);
    const asyncCallRef = React.useRef(new AsyncCall(executeFn, config, setState, null));
    const asyncCall = asyncCallRef.current;
    React.useEffect(() => {
      asyncCall.init();
      return () => {
        asyncCall.destroy$.next();
        asyncCall.destroy$.complete();
      };
    }, [asyncCall]);
    const status = asyncCall.newState.status;
    return {
      data: status.data,
      error: status.error,
      isLoading: status.status === 'PROCESSING',
      isSuccess: status.status === 'SUCCESS',
      execute: asyncCall.execute,
      reload: asyncCall.reload,
      status: asyncCall.getStatus(),
      getState: asyncCall.getState,
      isError: status.status === 'ERROR',
      setData: asyncCall.setData,
      ref: asyncCallRef,
      allErrors: asyncCall.newState.errors
    };
  };

  const useDetroy$ = () => {
    const destroy$Ref = React.useRef(new rxjs.Subject());
    const destroy$ = destroy$Ref.current;
    React.useEffect(() => {
      return () => {
        destroy$.next();
        destroy$.complete();
      };
    }, [destroy$]);
    return destroy$;
  };

  exports.EmitListenReturnState = void 0;

  (function (EmitListenReturnState) {
    EmitListenReturnState["OPEN"] = "OPEN";
    EmitListenReturnState["CLOSE"] = "CLOSE";
  })(exports.EmitListenReturnState || (exports.EmitListenReturnState = {}));

  const useEmitListen = (props = {}) => {
    const emitListen$Ref = React.useRef(new rxjs.Subject());
    const destroy$Ref = React.useRef(props.destroy$ ? props.destroy$ : new rxjs.Subject());
    const [state, setState] = React.useState(exports.EmitListenReturnState.OPEN);
    const open$Ref = React.useRef(new rxjs.Subject());
    const close$Ref = React.useRef(new rxjs.Subject());
    const destroy$ = destroy$Ref.current;
    React.useEffect(() => {
      open$Ref.current.pipe(rxjs.takeUntil(destroy$), rxjs.tap(() => {
        setState(exports.EmitListenReturnState.OPEN);
      })).subscribe();
      close$Ref.current.pipe(rxjs.takeUntil(destroy$), rxjs.tap(() => {
        setState(exports.EmitListenReturnState.CLOSE);
      })).subscribe();
      return () => {
        destroy$.next();
        destroy$.complete();
      };
    }, []);
    const params$ = props.params$ ? rxjs.merge(props.params$, emitListen$Ref.current) : emitListen$Ref.current;
    return {
      emit: emitListen$Ref.current.next.bind(emitListen$Ref.current),
      listen: params$.pipe(rxjs.takeUntil(destroy$Ref.current), rxjs.filter(x => state === exports.EmitListenReturnState.OPEN), rxjs.shareReplay()).subscribe.bind(emitListen$Ref.current),
      open: open$Ref.current.next.bind(open$Ref.current),
      close: close$Ref.current.next.bind(close$Ref.current),
      state
    };
  };

  function isEmptyInputValue(value) {
    return value == null || value.length === 0;
  }

  function hasValidLength(value) {
    return value != null && typeof value.length === 'number';
  }

  const EMAIL_REGEXP = /^(?=.{1,254}$)(?=.{1,64}@)[a-zA-Z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-zA-Z0-9!#$%&'*+/=?^_`{|}~-]+)*@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$/;
  class Validators {
    static min(min) {
      return minValidator(min);
    }

    static max(max) {
      return maxValidator(max);
    }

    static required(control) {
      return requiredValidator(control);
    }

    static requiredTrue(control) {
      return requiredTrueValidator(control);
    }

    static email(control) {
      return emailValidator(control);
    }

    static minLength(minLength) {
      return minLengthValidator(minLength);
    }

    static maxLength(maxLength) {
      return maxLengthValidator(maxLength);
    }

    static pattern(pattern) {
      return patternValidator(pattern);
    }

    static nullValidator(control) {
      return nullValidator();
    }

    static compose(validators) {
      return compose(validators);
    }

    static composeAsync(validators) {
      return composeAsync(validators);
    }

  }
  function minValidator(min) {
    return control => {
      if (isEmptyInputValue(control.value) || isEmptyInputValue(min)) {
        return null; // don't validate empty values to allow optional controls
      }

      const value = parseFloat(control.value);
      return !isNaN(value) && value < min ? {
        'min': {
          'min': min,
          'actual': control.value
        }
      } : null;
    };
  }
  function maxValidator(max) {
    return control => {
      if (isEmptyInputValue(control.value) || isEmptyInputValue(max)) {
        return null; // don't validate empty values to allow optional controls
      }

      const value = parseFloat(control.value);
      return !isNaN(value) && value > max ? {
        'max': {
          'max': max,
          'actual': control.value
        }
      } : null;
    };
  }
  function requiredValidator(control) {
    return isEmptyInputValue(control.value) ? {
      'required': true
    } : null;
  }
  function requiredTrueValidator(control) {
    return control.value === true ? null : {
      'required': true
    };
  }
  function emailValidator(control) {
    if (isEmptyInputValue(control.value)) {
      return null; // don't validate empty values to allow optional controls
    }

    return EMAIL_REGEXP.test(control.value) ? null : {
      'email': true
    };
  }
  function minLengthValidator(minLength) {
    return control => {
      if (isEmptyInputValue(control.value) || !hasValidLength(control.value)) {
        // don't validate empty values to allow optional controls
        // don't validate values without `length` property
        return null;
      }

      return control.value.length < minLength ? {
        'minlength': {
          'requiredLength': minLength,
          'actualLength': control.value.length
        }
      } : null;
    };
  }
  /**
   * Validator that requires the length of the control's value to be less than or equal
   * to the provided maximum length. See `Validators.maxLength` for additional information.
   */

  function maxLengthValidator(maxLength) {
    return control => {
      return hasValidLength(control.value) && control.value.length > maxLength ? {
        'maxlength': {
          'requiredLength': maxLength,
          'actualLength': control.value.length
        }
      } : null;
    };
  }
  /**
   * Validator that requires the control's value to match a regex pattern.
   * See `Validators.pattern` for additional information.
   */

  function patternValidator(pattern) {
    if (!pattern) return nullValidator;
    let regex;
    let regexStr;

    if (typeof pattern === 'string') {
      regexStr = '';
      if (pattern.charAt(0) !== '^') regexStr += '^';
      regexStr += pattern;
      if (pattern.charAt(pattern.length - 1) !== '$') regexStr += '$';
      regex = new RegExp(regexStr);
    } else {
      regexStr = pattern.toString();
      regex = pattern;
    }

    return control => {
      if (isEmptyInputValue(control.value)) {
        return null; // don't validate empty values to allow optional controls
      }

      const value = control.value;
      return regex.test(value) ? null : {
        'pattern': {
          'requiredPattern': regexStr,
          'actualValue': value
        }
      };
    };
  }
  /**
   * Function that has `ValidatorFn` shape, but performs no operation.
   */

  function nullValidator(control) {
    return null;
  }

  function isPresent(o) {
    return o != null;
  }

  function toObservable(r) {
    const obs = ModuleUtils.isPromise(r) ? rxjs.from(r) : r;

    if (!ModuleUtils.isObservable(obs)
    /* && (typeof ngDevMode === 'undefined' || ngDevMode)*/
    ) {
      throw new Error(`Expected validator to return Promise or Observable.`);
    }

    return obs;
  }

  function mergeErrors(arrayOfErrors) {
    let res = {}; // Not using Array.reduce here due to a Chrome 80 bug
    // https://bugs.chromium.org/p/chromium/issues/detail?id=1049982

    arrayOfErrors.forEach(errors => {
      res = errors != null ? _extends({}, res, errors) : res;
    });
    return Object.keys(res).length === 0 ? null : res;
  }

  function executeValidators(control, validators) {
    return validators.map(validator => validator(control));
  }

  function isValidatorFn(validator) {
    return !validator.validate;
  }
  /**
   * Given the list of validators that may contain both functions as well as classes, return the list
   * of validator functions (convert validator classes into validator functions). This is needed to
   * have consistent structure in validators list before composing them.
   *
   * @param validators The set of validators that may contain validators both in plain function form
   *     as well as represented as a validator class.
   */


  function normalizeValidators(validators) {
    return validators.map(validator => {
      return isValidatorFn(validator) ? validator : c => validator.validate(c);
    });
  }
  /**
   * Merges synchronous validators into a single validator function.
   * See `Validators.compose` for additional information.
   */

  function compose(validators) {
    if (!validators) return null;
    const presentValidators = validators.filter(isPresent);
    if (presentValidators.length == 0) return null;
    return function (control) {
      return mergeErrors(executeValidators(control, presentValidators));
    };
  }
  /**
   * Accepts a list of validators of different possible shapes (`Validator` and `ValidatorFn`),
   * normalizes the list (converts everything to `ValidatorFn`) and merges them into a single
   * validator function.
   */


  function composeValidators(validators) {
    return validators != null ? compose(normalizeValidators(validators)) : null;
  }
  /**
   * Merges asynchronous validators into a single validator function.
   * See `Validators.composeAsync` for additional information.
   */

  function composeAsync(validators) {
    if (!validators) return null;
    const presentValidators = validators.filter(isPresent);
    if (presentValidators.length == 0) return null;
    return function (control) {
      const observables = executeValidators(control, presentValidators).map(toObservable);
      return rxjs.forkJoin(observables).pipe(map(mergeErrors));
    };
  }
  /**
   * Accepts a list of async validators of different possible shapes (`AsyncValidator` and
   * `AsyncValidatorFn`), normalizes the list (converts everything to `AsyncValidatorFn`) and merges
   * them into a single validator function.
   */


  function composeAsyncValidators(validators) {
    return validators != null ? composeAsync(normalizeValidators(validators)) : null;
  }
  /**
   * Merges raw control validators with a given directive validator and returns the combined list of
   * validators as an array.
   */

  function mergeValidators(controlValidators, dirValidator) {
    if (controlValidators === null) return [dirValidator];
    return Array.isArray(controlValidators) ? [...controlValidators, dirValidator] : [controlValidators, dirValidator];
  }
  /**
   * Retrieves the list of raw synchronous validators attached to a given control.
   */

  function getControlValidators(control) {
    return control._rawValidators;
  }
  /**
   * Retrieves the list of raw asynchronous validators attached to a given control.
   */

  function getControlAsyncValidators(control) {
    return control._rawAsyncValidators;
  }
  /**
   * Accepts a singleton validator, an array, or null, and returns an array type with the provided
   * validators.
   *
   * @param validators A validator, validators, or null.
   * @returns A validators array.
   */

  function makeValidatorsArray(validators) {
    if (!validators) return [];
    return Array.isArray(validators) ? validators : [validators];
  }
  /**
   * Determines whether a validator or validators array has a given validator.
   *
   * @param validators The validator or validators to compare against.
   * @param validator The validator to check.
   * @returns Whether the validator is present.
   */

  function hasValidator(validators, validator) {
    return Array.isArray(validators) ? validators.includes(validator) : validators === validator;
  }
  /**
   * Combines two arrays of validators into one. If duplicates are provided, only one will be added.
   *
   * @param validators The new validators.
   * @param currentValidators The base array of currrent validators.
   * @returns An array of validators.
   */

  function addValidators(validators, currentValidators) {
    const current = makeValidatorsArray(currentValidators);
    const validatorsToAdd = makeValidatorsArray(validators);
    validatorsToAdd.forEach(v => {
      // Note: if there are duplicate entries in the new validators array,
      // only the first one would be added to the current list of validarors.
      // Duplicate ones would be ignored since `hasValidator` would detect
      // the presence of a validator function and we update the current list in place.
      if (!hasValidator(current, v)) {
        current.push(v);
      }
    });
    return current;
  }
  function removeValidators(validators, currentValidators) {
    return makeValidatorsArray(currentValidators).filter(v => !hasValidator(validators, v));
  }

  class EventEmitter_ extends rxjs.Subject {
    // tslint:disable-line
    constructor(isAsync = false) {
      super();
      this.__isAsync = void 0;
      this.__isAsync = isAsync;
    }

    emit(value) {
      super.next(value);
    }

    subscribe(observerOrNext, error, complete) {
      let nextFn = observerOrNext;

      let errorFn = error || (() => null);

      let completeFn = complete;

      if (observerOrNext && typeof observerOrNext === 'object') {
        var _observer$next, _observer$error, _observer$complete;

        const observer = observerOrNext;
        nextFn = (_observer$next = observer.next) == null ? void 0 : _observer$next.bind(observer);
        errorFn = (_observer$error = observer.error) == null ? void 0 : _observer$error.bind(observer);
        completeFn = (_observer$complete = observer.complete) == null ? void 0 : _observer$complete.bind(observer);
      }

      if (this.__isAsync) {
        errorFn = _wrapInTimeout(errorFn);

        if (nextFn) {
          nextFn = _wrapInTimeout(nextFn);
        }

        if (completeFn) {
          completeFn = _wrapInTimeout(completeFn);
        }
      }

      const sink = super.subscribe({
        next: nextFn,
        error: errorFn,
        complete: completeFn
      });

      if (observerOrNext instanceof rxjs.Subscription) {
        observerOrNext.add(sink);
      }

      return sink;
    }

  }

  function _wrapInTimeout(fn) {
    return value => {
      setTimeout(fn, undefined, value);
    };
  }
  /**
   * @publicApi
   */


  const EventEmitter2 = EventEmitter_;

  const ARRAY_WILDCARD = "+";
  const PATH_DELIMITER = ".";

  class ObjectPrototypeMutationError extends Error {
    constructor(params) {
      super(params);
      this.name = "ObjectPrototypeMutationError";
    }

  }

  const getValue = (object, property) => {
    if (typeof object != "object" || object === null) {
      return object;
    }

    if (typeof property == "undefined") {
      return object;
    }

    if (typeof property == "number") {
      return object[property];
    }

    try {
      return traverse(object, property, (currentObject, currentProperty) => {
        return currentObject[currentProperty];
      });
    } catch (err) {
      return object;
    }
  };
  const setValue = (object, property, value) => {
    if (typeof object != "object" || object === null) {
      return object;
    }

    if (typeof property == "undefined") {
      return object;
    }

    if (typeof property == "number") {
      object[property] = value;
      return object[property];
    }

    try {
      return traverse(object, property, (currentObject, currentProperty, segments, index) => {
        if (currentObject === Reflect.getPrototypeOf({})) {
          throw new ObjectPrototypeMutationError("Attempting to mutate Object.prototype");
        }

        if (!currentObject[currentProperty]) {
          const nextPropIsNumber = Number.isInteger(Number(segments[index + 1]));
          const nextPropIsArrayWildcard = segments[index + 1] === ARRAY_WILDCARD;

          if (nextPropIsNumber || nextPropIsArrayWildcard) {
            currentObject[currentProperty] = [];
          } else {
            currentObject[currentProperty] = {};
          }
        }

        if (isLastSegment(segments, index)) {
          currentObject[currentProperty] = value;
        }

        return currentObject[currentProperty];
      });
    } catch (err) {
      if (err instanceof ObjectPrototypeMutationError) {
        throw err;
      } else {
        return object;
      }
    }
  };

  const traverse = (object, path, callback = () => {}) => {
    const segments = path.split(PATH_DELIMITER);
    const length = segments.length;

    for (let idx = 0; idx < length; idx++) {
      const currentSegment = segments[idx];

      if (!object) {
        return;
      }

      if (currentSegment === ARRAY_WILDCARD) {
        if (Array.isArray(object)) {
          return object.map((value, index) => {
            const remainingSegments = segments.slice(idx + 1);

            if (remainingSegments.length > 0) {
              return traverse(value, remainingSegments.join(PATH_DELIMITER), callback);
            } else {
              return callback(object, index, segments, idx);
            }
          });
        } else {
          const pathToHere = segments.slice(0, idx).join(PATH_DELIMITER);
          throw new Error(`Object at wildcard (${pathToHere}) is not an array`);
        }
      } else {
        object = callback(object, currentSegment, segments, idx);
      }
    }

    return object;
  };

  function isLastSegment(segments, index) {
    return segments.length === index + 1;
  }

  function removeListItem(list, el) {
    const index = list.indexOf(el);
    if (index > -1) list.splice(index, 1);
  }

  const VALID = 'VALID';
  const INVALID = 'INVALID';
  const PENDING = 'PENDING';
  const DISABLED = 'DISABLED';

  function _find(control, path, delimiter) {
    if (path == null) return null;

    if (!Array.isArray(path)) {
      path = path.split(delimiter);
    }

    if (Array.isArray(path) && path.length === 0) return null;
    let controlToFind = control;
    path.forEach(name => {
      if (controlToFind instanceof FormGroup) {
        controlToFind = controlToFind.controls.hasOwnProperty(name) ? controlToFind.controls[name] : null;
      } else if (controlToFind instanceof FormArray) {
        controlToFind = controlToFind.at(name) || null;
      } else {
        controlToFind = null;
      }
    });
    return controlToFind;
  }

  function pickValidators(validatorOrOpts) {
    return (isOptionsObj(validatorOrOpts) ? validatorOrOpts.validators : validatorOrOpts) || null;
  }

  function coerceToValidator(validator) {
    return Array.isArray(validator) ? composeValidators(validator) : validator || null;
  }

  function pickAsyncValidators(asyncValidator, validatorOrOpts) {
    return (isOptionsObj(validatorOrOpts) ? validatorOrOpts.asyncValidators : asyncValidator) || null;
  }

  function coerceToAsyncValidator(asyncValidator) {
    return Array.isArray(asyncValidator) ? composeAsyncValidators(asyncValidator) : asyncValidator || null;
  }

  function isOptionsObj(validatorOrOpts) {
    return validatorOrOpts != null && !Array.isArray(validatorOrOpts) && typeof validatorOrOpts === 'object';
  }

  class AbstractControl {
    constructor(validators, asyncValidators) {
      this._pendingDirty = void 0;
      this._hasOwnPendingAsyncValidator = false;
      this._pendingTouched = void 0;

      this._onCollectionChange = () => {};

      this._updateOn = void 0;
      this._parent = null;
      this._asyncValidationSubscription = void 0;
      this._composedValidatorFn = void 0;
      this._composedAsyncValidatorFn = void 0;
      this._rawValidators = void 0;
      this._rawAsyncValidators = void 0;
      this.value = void 0;
      this.status = void 0;
      this.errors = void 0;
      this.pristine = true;
      this.touched = false;
      this.valueChanges = void 0;
      this.statusChanges = void 0;
      this._onDisabledChange = [];
      this.nestedPath = void 0;
      this.path = null;

      this.dispatchStateFn = vale => {};

      this._rawValidators = validators;
      this._rawAsyncValidators = asyncValidators;
      this._composedValidatorFn = coerceToValidator(this._rawValidators);
      this._composedAsyncValidatorFn = coerceToAsyncValidator(this._rawAsyncValidators);
    }

    get validator() {
      return this._composedValidatorFn;
    }

    set validator(validatorFn) {
      this._rawValidators = this._composedValidatorFn = validatorFn;
    }

    get asyncValidator() {
      return this._composedAsyncValidatorFn;
    }

    set asyncValidator(asyncValidatorFn) {
      this._rawAsyncValidators = this._composedAsyncValidatorFn = asyncValidatorFn;
    }

    get parent() {
      return this._parent;
    }

    get valid() {
      return this.status === VALID;
    }

    get invalid() {
      return this.status === INVALID;
    }

    get pending() {
      return this.status == PENDING;
    }

    get disabled() {
      return this.status === DISABLED;
    }

    get enabled() {
      return this.status !== DISABLED;
    }

    get dirty() {
      return !this.pristine;
    }

    get untouched() {
      return !this.touched;
    }

    get updateOn() {
      return this._updateOn ? this._updateOn : this.parent ? this.parent.updateOn : 'change';
    }

    setValidators(validators) {
      this._rawValidators = validators;
      this._composedValidatorFn = coerceToValidator(validators);
    }

    setAsyncValidators(validators) {
      this._rawAsyncValidators = validators;
      this._composedAsyncValidatorFn = coerceToAsyncValidator(validators);
    }

    addValidators(validators) {
      this.setValidators(addValidators(validators, this._rawValidators));
    }

    addAsyncValidators(validators) {
      this.setAsyncValidators(addValidators(validators, this._rawAsyncValidators));
    }

    removeValidators(validators) {
      this.setValidators(removeValidators(validators, this._rawValidators));
    }

    removeAsyncValidators(validators) {
      this.setAsyncValidators(removeValidators(validators, this._rawAsyncValidators));
    }

    hasValidator(validator) {
      return hasValidator(this._rawValidators, validator);
    }

    hasAsyncValidator(validator) {
      return hasValidator(this._rawAsyncValidators, validator);
    }

    clearValidators() {
      this.validator = null;
    }

    clearAsyncValidators() {
      this.asyncValidator = null;
    }

    markAsTouched(opts = {}) {
      this.touched = true;

      if (this._parent && !opts.onlySelf) {
        this._parent.markAsTouched(opts);
      }
    }

    markAllAsTouched() {
      this.markAsTouched({
        onlySelf: true
      });

      this._forEachChild(control => control.markAllAsTouched());
    }

    markAsUntouched(opts = {}) {
      this.touched = false;
      this._pendingTouched = false;

      this._forEachChild(control => {
        control.markAsUntouched({
          onlySelf: true
        });
      });

      if (this._parent && !opts.onlySelf) {
        this._parent._updateTouched(opts);
      }
    }

    markAsDirty(opts = {}) {
      this.pristine = false;

      if (this._parent && !opts.onlySelf) {
        this._parent.markAsDirty(opts);
      }
    }

    markAsPristine(opts = {}) {
      this.pristine = true;
      this._pendingDirty = false;

      this._forEachChild(control => {
        control.markAsPristine({
          onlySelf: true
        });
      });

      if (this._parent && !opts.onlySelf) {
        this._parent._updatePristine(opts);
      }
    }

    markAsPending(opts = {}) {
      this.status = PENDING;

      if (opts.emitEvent !== false) {
        this.statusChanges.emit(this.status);
      }

      if (this._parent && !opts.onlySelf) {
        this._parent.markAsPending(opts);
      }
    }

    disable(opts = {}) {
      const skipPristineCheck = this._parentMarkedDirty(opts.onlySelf);

      this.status = DISABLED;
      this.errors = null;

      this._forEachChild(control => {
        control.disable(_extends({}, opts, {
          onlySelf: true
        }));
      });

      this._updateValue();

      if (opts.emitEvent !== false) {
        this.valueChanges.emit(this.value);
        this.statusChanges.emit(this.status);
      }

      this._updateAncestors(_extends({}, opts, {
        skipPristineCheck
      }));

      this._onDisabledChange.forEach(changeFn => changeFn(true));
    }

    enable(opts = {}) {
      // If parent has been marked artificially dirty we don't want to re-calculate the
      // parent's dirtiness based on the children.
      const skipPristineCheck = this._parentMarkedDirty(opts.onlySelf);

      this.status = VALID;

      this._forEachChild(control => {
        control.enable(_extends({}, opts, {
          onlySelf: true
        }));
      });

      this.updateValueAndValidity({
        onlySelf: true,
        emitEvent: opts.emitEvent
      });

      this._updateAncestors(_extends({}, opts, {
        skipPristineCheck
      }));

      this._onDisabledChange.forEach(changeFn => changeFn(false));
    }

    _updateAncestors(opts) {
      if (this._parent && !opts.onlySelf) {
        this._parent.updateValueAndValidity(opts);

        if (!opts.skipPristineCheck) {
          this._parent._updatePristine();
        }

        this._parent._updateTouched();
      }
    }

    setParent(parent) {
      this._parent = parent;
    }

    updateValueAndValidity(opts = {}) {
      this._setInitialStatus();

      this._updateValue();

      if (this.enabled) {
        this._cancelExistingSubscription();

        this.errors = this._runValidator();
        this.status = this._calculateStatus();

        if (this.status === VALID || this.status === PENDING) {
          this._runAsyncValidator(opts.emitEvent);
        }
      }

      if (opts.emitEvent !== false) {
        this.valueChanges.emit(this.value);
        this.statusChanges.emit(this.status);
      }

      if (this._parent && !opts.onlySelf) {
        this._parent.updateValueAndValidity(opts);
      }
    }

    _updateTreeValidity(opts = {
      emitEvent: true
    }) {
      this._forEachChild(ctrl => ctrl._updateTreeValidity(opts));

      this.updateValueAndValidity({
        onlySelf: true,
        emitEvent: opts.emitEvent
      });
    }

    _setInitialStatus() {
      this.status = this._allControlsDisabled() ? DISABLED : VALID;
    }

    _runValidator() {
      return this.validator ? this.validator(this) : null;
    }

    _runAsyncValidator(emitEvent) {
      if (this.asyncValidator) {
        this.status = PENDING;
        this._hasOwnPendingAsyncValidator = true;
        const obs = toObservable(this.asyncValidator(this));
        this._asyncValidationSubscription = obs.subscribe(errors => {
          this._hasOwnPendingAsyncValidator = false;
          this.setErrors(errors, {
            emitEvent
          });
        });
      }
    }

    _cancelExistingSubscription() {
      if (this._asyncValidationSubscription) {
        this._asyncValidationSubscription.unsubscribe();

        this._hasOwnPendingAsyncValidator = false;
      }
    }

    setErrors(errors, opts = {}) {
      this.errors = errors;

      this._updateControlsErrors(opts.emitEvent !== false);
    }

    get(path) {
      return _find(this, path, '.');
    }

    getError(errorCode, path) {
      const control = path ? this.get(path) : this;
      return control && control.errors ? control.errors[errorCode] : null;
    }

    hasError(errorCode, path) {
      return !!this.getError(errorCode, path);
    }

    get root() {
      let x = this;

      while (x._parent) {
        x = x._parent;
      }

      return x;
    }
    /** @internal */


    _updateControlsErrors(emitEvent) {
      this.status = this._calculateStatus();

      if (emitEvent) {
        this.statusChanges.emit(this.status);
      }

      if (this._parent) {
        this._parent._updateControlsErrors(emitEvent);
      }
    }
    /** @internal */


    _initObservables() {
      this.valueChanges = new EventEmitter2();
      this.statusChanges = new EventEmitter2();
    }

    _calculateStatus() {
      if (this._allControlsDisabled()) return DISABLED;
      if (this.errors) return INVALID;
      if (this._hasOwnPendingAsyncValidator || this._anyControlsHaveStatus(PENDING)) return PENDING;
      if (this._anyControlsHaveStatus(INVALID)) return INVALID;
      return VALID;
    }
    /** @internal */


    /** @internal */
    _anyControlsHaveStatus(status) {
      return this._anyControls(control => control.status === status);
    }
    /** @internal */


    _anyControlsDirty() {
      return this._anyControls(control => control.dirty);
    }
    /** @internal */


    _anyControlsTouched() {
      return this._anyControls(control => control.touched);
    }
    /** @internal */


    _updatePristine(opts = {}) {
      this.pristine = !this._anyControlsDirty();

      if (this._parent && !opts.onlySelf) {
        this._parent._updatePristine(opts);
      }
    }
    /** @internal */


    _updateTouched(opts = {}) {
      this.touched = this._anyControlsTouched();

      if (this._parent && !opts.onlySelf) {
        this._parent._updateTouched(opts);
      }
    }
    /** @internal */


    /** @internal */
    _isBoxedValue(formState) {
      return typeof formState === 'object' && formState !== null && Object.keys(formState).length === 2 && 'value' in formState && 'disabled' in formState;
    }
    /** @internal */


    _registerOnCollectionChange(fn) {
      this._onCollectionChange = fn;
    }
    /** @internal */


    _setUpdateStrategy(opts) {
      if (isOptionsObj(opts) && opts.updateOn != null) {
        this._updateOn = opts.updateOn;
      }
    }

    _parentMarkedDirty(onlySelf) {
      const parentDirty = this._parent && this._parent.dirty;
      return !onlySelf && !!parentDirty && !this._parent._anyControlsDirty();
    }

    getRoot() {
      if (this.parent) {
        return this.parent.getRoot();
      } else {
        return this;
      }
    }

    setState(value) {
      const root = this.getRoot();
      const dispatchStateFn = root.dispatchStateFn;
      const path = (root.nestedPath ? [root.nestedPath, ...this.getPath()] : this.getPath()).join('.');

      if (path === '') {
        dispatchStateFn(value);
      } else {
        const newState = produce__default["default"](draftState => {
          setValue(draftState, path, value);
        });
        dispatchStateFn(newState);
      }
    }

    getPath() {
      if (this.parent) return [...this.parent.getPath(), this.path];else return [];
    }

    getPathString() {
      return this.getPath().join('.');
    }

  }
  class FormControl extends AbstractControl {
    constructor(formState = null, validatorOrOpts, asyncValidator) {
      super(pickValidators(validatorOrOpts), pickAsyncValidators(asyncValidator, validatorOrOpts));
      this._onChange = [];
      this._pendingValue = void 0;
      this._pendingChange = void 0;

      this.onChange = event => {
        const value = getControlValue(event);
        const isDirty = value !== this.value;

        if (this.updateOn !== "change") {
          this._pendingValue = value;
          this._pendingChange = true;

          if (isDirty && !this._pendingDirty) {
            this._pendingDirty = true;
          }
        } else {
          if (isDirty && !this.dirty) {
            this.markAsDirty();
          }

          this.setValue(value);
        }
      };

      this.onBlur = () => {
        if (this.updateOn === "blur") {
          if (this._pendingDirty && !this.dirty) {
            this.markAsDirty();
          }

          if (!this.touched) {
            this.markAsTouched();
          }

          this.setValue(this._pendingValue, {
            emitEvent: false,
            onlySelf: true
          });
        } else if (this.updateOn === "submit") {
          this._pendingTouched = true;
        } else {
          !this.touched;

          if (!this.touched) {
            this.markAsTouched();
            this.setValue(this.value);
          }
        }
      };

      this._applyFormState(formState);

      this._setUpdateStrategy(validatorOrOpts);

      this._initObservables();

      this.updateValueAndValidity({
        onlySelf: true,
        emitEvent: !!this.asyncValidator
      });
    }

    setValue(value, options = {}) {
      this.value = this._pendingValue = value;
      const root = this.getRoot();
      const tempPath = this.getPath().join('.');
      root.nestedPath ? `${root.nestedPath}.${tempPath}` : tempPath;
      this.setState(value);

      if (this._onChange.length && options.emitModelToViewChange !== false) {
        this._onChange.forEach(changeFn => changeFn(this.value, options.emitViewToModelChange !== false));
      }

      this.updateValueAndValidity(options);
    }

    patchValue(value, options = {}) {
      this.setValue(value, options);
    }

    reset(formState = null, options = {}) {
      this._applyFormState(formState);

      this.markAsPristine(options);
      this.markAsUntouched(options);
      this.setValue(this.value, options);
      this._pendingChange = false;
    }

    _updateValue() {}

    _anyControls(condition) {
      return false;
    }

    _allControlsDisabled() {
      return this.disabled;
    }

    registerOnChange(fn) {
      this._onChange.push(fn);
    }

    _unregisterOnChange(fn) {
      removeListItem(this._onChange, fn);
    }

    registerOnDisabledChange(fn) {
      this._onDisabledChange.push(fn);
    }

    _unregisterOnDisabledChange(fn) {
      removeListItem(this._onDisabledChange, fn);
    }

    _forEachChild(cb) {}

    _syncPendingControls() {
      if (this.updateOn === 'submit') {
        if (this._pendingDirty) this.markAsDirty();
        if (this._pendingTouched) this.markAsTouched();

        if (this._pendingChange) {
          this.setValue(this._pendingValue, {
            onlySelf: true,
            emitModelToViewChange: false
          });
          return true;
        }
      }

      return false;
    }

    _applyFormState(formState) {
      if (this._isBoxedValue(formState)) {
        this.value = this._pendingValue = formState.value;
        formState.disabled ? this.disable({
          onlySelf: true,
          emitEvent: false
        }) : this.enable({
          onlySelf: true,
          emitEvent: false
        });
      } else {
        this.value = this._pendingValue = formState;
      }
    }

  }
  class FormGroup extends AbstractControl {
    constructor(controls, validatorOrOpts, asyncValidator) {
      super(pickValidators(validatorOrOpts), pickAsyncValidators(asyncValidator, validatorOrOpts));
      this.controls = controls;

      this._initObservables();

      this._setUpdateStrategy(validatorOrOpts);

      this._setUpControls();

      this.updateValueAndValidity({
        onlySelf: true,
        emitEvent: !!this.asyncValidator
      });
    }

    registerControl(name, control) {
      if (this.controls[name]) return this.controls[name];
      this.controls[name] = control;
      control.setParent(this);

      control._registerOnCollectionChange(this._onCollectionChange);

      return control;
    }

    addControl(name, control, options = {}) {
      this.registerControl(name, control);
      this.updateValueAndValidity({
        emitEvent: options.emitEvent
      });

      this._onCollectionChange();
    }

    removeControl(name, options = {}) {
      if (this.controls[name]) this.controls[name]._registerOnCollectionChange(() => {});
      delete this.controls[name];
      this.updateValueAndValidity({
        emitEvent: options.emitEvent
      });

      this._onCollectionChange();
    }

    setControl(name, control, options = {}) {
      if (this.controls[name]) this.controls[name]._registerOnCollectionChange(() => {});
      delete this.controls[name];
      if (control) this.registerControl(name, control);
      this.updateValueAndValidity({
        emitEvent: options.emitEvent
      });

      this._onCollectionChange();
    }

    contains(controlName) {
      return this.controls.hasOwnProperty(controlName) && this.controls[controlName].enabled;
    }

    setValue(value, options = {}) {
      this._checkAllValuesPresent(value);

      Object.keys(value).forEach(name => {
        this._throwIfControlMissing(name);

        this.controls[name].setValue(value[name], {
          onlySelf: true,
          emitEvent: options.emitEvent
        });
      });
      this.updateValueAndValidity(options);
    }

    patchValue(value, options = {}) {
      if (value == null
      /* both `null` and `undefined` */
      ) return;
      Object.keys(value).forEach(name => {
        if (this.controls[name]) {
          this.controls[name].patchValue(value[name], {
            onlySelf: true,
            emitEvent: options.emitEvent
          });
        }
      });
      this.updateValueAndValidity(options);
      this.setState(value);
    }

    reset(value = {}, options = {}) {
      this._forEachChild((control, name) => {
        control.reset(value[name], {
          onlySelf: true,
          emitEvent: options.emitEvent
        });
      });

      this._updatePristine(options);

      this._updateTouched(options);

      this.updateValueAndValidity(options);
    }

    getRawValue() {
      return this._reduceChildren({}, (acc, control, name) => {
        acc[name] = control instanceof FormControl ? control.value : control.getRawValue();
        return acc;
      });
    }

    _syncPendingControls() {
      let subtreeUpdated = this._reduceChildren(false, (updated, child) => {
        return child._syncPendingControls() ? true : updated;
      });

      if (subtreeUpdated) this.updateValueAndValidity({
        onlySelf: true
      });
      return subtreeUpdated;
    }

    _throwIfControlMissing(name) {
      if (!Object.keys(this.controls).length) {
        throw new Error(`
         There are no form controls registered with this group yet. If you're using ngModel,
         you may want to check next tick (e.g. use setTimeout).
       `);
      }

      if (!this.controls[name]) {
        throw new Error(`Cannot find form control with name: ${name}.`);
      }
    }

    _forEachChild(cb) {
      Object.keys(this.controls).forEach(key => {
        const control = this.controls[key];
        control && cb(control, key);
      });
    }

    _setUpControls() {
      this._forEachChild((control, key) => {
        control.setParent(this);
        control.path = key;

        control._registerOnCollectionChange(this._onCollectionChange);
      });
    }

    _updateValue() {
      this.value = this._reduceValue();
    }

    _anyControls(condition) {
      for (const controlName of Object.keys(this.controls)) {
        const control = this.controls[controlName];

        if (this.contains(controlName) && condition(control)) {
          return true;
        }
      }

      return false;
    }

    _reduceValue() {
      return this._reduceChildren({}, (acc, control, name) => {
        if (control.enabled || this.disabled) {
          acc[name] = control.value;
        }

        return acc;
      });
    }
    /** @internal */


    _reduceChildren(initValue, fn) {
      let res = initValue;

      this._forEachChild((control, name) => {
        res = fn(res, control, name);
      });

      return res;
    }

    _allControlsDisabled() {
      for (const controlName of Object.keys(this.controls)) {
        if (this.controls[controlName].enabled) {
          return false;
        }
      }

      return Object.keys(this.controls).length > 0 || this.disabled;
    }
    /** @internal */


    _checkAllValuesPresent(value) {
      this._forEachChild((control, name) => {
        if (value[name] === undefined) {
          throw new Error(`Must supply a value for form control with name: '${name}'.`);
        }
      });
    }

  }
  class FormArray extends AbstractControl {
    constructor(controls, validatorOrOpts, asyncValidator) {
      super(pickValidators(validatorOrOpts), pickAsyncValidators(asyncValidator, validatorOrOpts));
      this.controls = controls;

      this._initObservables();

      this._setUpdateStrategy(validatorOrOpts);

      this._setUpControls();

      this.updateValueAndValidity({
        onlySelf: true,
        emitEvent: !!this.asyncValidator
      });
    }

    at(index) {
      return this.controls[index];
    }

    push(control, options = {}) {
      this.controls.push(control);
      control.path = this.controls.length - 1;
      const path = this.getPath().join('.') + '.' + control.path;
      const newState = produce__default["default"](draftState => {
        setValue(draftState, path, control.value);
      });
      this.setState(newState);

      this._registerControl(control);

      this.updateValueAndValidity({
        emitEvent: options.emitEvent
      });

      this._onCollectionChange();
    }

    insert(index, control, options = {}) {
      this.controls.splice(index, 0, control);

      this._registerControl(control);

      this.updateValueAndValidity({
        emitEvent: options.emitEvent
      });
    }

    removeAt(index, options = {}) {
      if (this.controls[index]) this.controls[index]._registerOnCollectionChange(() => {});
      const path = this.getPath().join('.');
      const newState = produce__default["default"](draftState => {
        const controls = getValue(draftState, path);
        controls.splice(index, 1);
      });
      this.setState(newState);
      this.controls.forEach((control, index) => {
        control.path = index;
      });
      this.controls.splice(index, 1);
      this.updateValueAndValidity({
        emitEvent: options.emitEvent
      });
    }

    setControl(index, control, options = {}) {
      if (this.controls[index]) this.controls[index]._registerOnCollectionChange(() => {});
      this.controls.splice(index, 1);

      if (control) {
        this.controls.splice(index, 0, control);

        this._registerControl(control);
      }

      this.updateValueAndValidity({
        emitEvent: options.emitEvent
      });

      this._onCollectionChange();
    }

    get length() {
      return this.controls.length;
    }

    setValue(value, options = {}) {
      this._checkAllValuesPresent(value);

      value.forEach((newValue, index) => {
        this._throwIfControlMissing(index);

        this.at(index).setValue(newValue, {
          onlySelf: true,
          emitEvent: options.emitEvent
        });
      });
      this.updateValueAndValidity(options);
    }

    patchValue(value, options = {}) {
      if (value == null
      /* both `null` and `undefined` */
      ) return;
      value.forEach((newValue, index) => {
        if (this.at(index)) {
          this.at(index).patchValue(newValue, {
            onlySelf: true,
            emitEvent: options.emitEvent
          });
        }
      });
      this.updateValueAndValidity(options);
    }

    reset(value = [], options = {}) {
      this._forEachChild((control, index) => {
        control.reset(value[index], {
          onlySelf: true,
          emitEvent: options.emitEvent
        });
      });

      this._updatePristine(options);

      this._updateTouched(options);

      this.updateValueAndValidity(options);
    }

    getRawValue() {
      return this.controls.map(control => {
        return control instanceof FormControl ? control.value : control.getRawValue();
      });
    }

    clear(options = {}) {
      if (this.controls.length < 1) return;

      this._forEachChild(control => control._registerOnCollectionChange(() => {}));

      this.controls.splice(0);
      this.updateValueAndValidity({
        emitEvent: options.emitEvent
      });
    }

    _syncPendingControls() {
      let subtreeUpdated = this.controls.reduce((updated, child) => {
        return child._syncPendingControls() ? true : updated;
      }, false);
      if (subtreeUpdated) this.updateValueAndValidity({
        onlySelf: true
      });
      return subtreeUpdated;
    }

    _throwIfControlMissing(index) {
      if (!this.controls.length) {
        throw new Error(`
         There are no form controls registered with this array yet. If you're using ngModel,
         you may want to check next tick (e.g. use setTimeout).
       `);
      }

      if (!this.at(index)) {
        throw new Error(`Cannot find form control at index ${index}`);
      }
    }

    _forEachChild(cb) {
      this.controls.forEach((control, index) => {
        cb(control, index);
      });
    }

    _updateValue() {
      this.value = this.controls.filter(control => control.enabled || this.disabled).map(control => control.value);
    }

    _anyControls(condition) {
      return this.controls.some(control => control.enabled && condition(control));
    }

    _setUpControls() {
      this._forEachChild((control, index) => {
        this._registerControl(control);

        control.path = index;
      });
    }

    _checkAllValuesPresent(value) {
      this._forEachChild((control, i) => {
        if (value[i] === undefined) {
          throw new Error(`Must supply a value for form control at index: ${i}.`);
        }
      });
    }

    _allControlsDisabled() {
      for (const control of this.controls) {
        if (control.enabled) return false;
      }

      return this.controls.length > 0 || this.disabled;
    }

    _registerControl(control) {
      control.setParent(this);

      control._registerOnCollectionChange(this._onCollectionChange);
    }

  }

  function getControlValue(event) {
    if (isEvent(event)) {
      switch (event.target.type) {
        case "checkbox":
          return event.target.checked;

        case "select-multiple":
          if (event.target.options) {
            let options = event.target.options;
            var value = [];

            for (var i = 0, l = options.length; i < l; i++) {
              if (options[i].selected) {
                value.push(options[i].value);
              }
            }

            return value;
          }

          return event.target.value;

        case "number":
          return event.target.valueAsNumber;

        default:
          return isReactNative() ? event.nativeEvent.text : event.target.value;
      }
    }

    return event;
  }

  const isReactNative = () => typeof window !== "undefined" && window.navigator && window.navigator.product && window.navigator.product === "ReactNative";
  const isEvent = candidate => !!(candidate && candidate.stopPropagation && candidate.preventDefault);

  function isAbstractControlOptions(options) {
    return options.asyncValidators !== undefined || options.validators !== undefined || options.updateOn !== undefined;
  }
  /**
   * @description
   * Creates an `AbstractControl` from a user-specified configuration.
   *
   * The `FormBuilder` provides syntactic sugar that shortens creating instances of a `FormControl`,
   * `FormGroup`, or `FormArray`. It reduces the amount of boilerplate needed to build complex
   * forms.
   *
   * @see [Reactive Forms Guide](/guide/reactive-forms)
   *
   * @publicApi
   */
  //  @Injectable({providedIn: ReactiveFormsModule})


  class FormBuilder {
    /**
     * @description
     * Construct a new `FormGroup` instance.
     *
     * @param controlsConfig A collection of child controls. The key for each child is the name
     * under which it is registered.
     *
     * @param options Configuration options object for the `FormGroup`. The object should have the
     * the `AbstractControlOptions` type and might contain the following fields:
     * * `validators`: A synchronous validator function, or an array of validator functions
     * * `asyncValidators`: A single async validator or array of async validator functions
     * * `updateOn`: The event upon which the control should be updated (options: 'change' | 'blur' |
     * submit')
     */

    /**
     * @description
     * Construct a new `FormGroup` instance.
     *
     * @deprecated This API is not typesafe and can result in issues with Closure Compiler renaming.
     * Use the `FormBuilder#group` overload with `AbstractControlOptions` instead.
     * Note that `AbstractControlOptions` expects `validators` and `asyncValidators` to be valid
     * validators. If you have custom validators, make sure their validation function parameter is
     * `AbstractControl` and not a sub-class, such as `FormGroup`. These functions will be called with
     * an object of type `AbstractControl` and that cannot be automatically downcast to a subclass, so
     * TypeScript sees this as an error. For example, change the `(group: FormGroup) =>
     * ValidationErrors|null` signature to be `(group: AbstractControl) => ValidationErrors|null`.
     *
     * @param controlsConfig A collection of child controls. The key for each child is the name
     * under which it is registered.
     *
     * @param options Configuration options object for the `FormGroup`. The legacy configuration
     * object consists of:
     * * `validator`: A synchronous validator function, or an array of validator functions
     * * `asyncValidator`: A single async validator or array of async validator functions
     * Note: the legacy format is deprecated and might be removed in one of the next major versions
     * of Angular.
     */
    group(controlsConfig, options = null) {
      const controls = this._reduceControls(controlsConfig);

      let validators = null;
      let asyncValidators = null;
      let updateOn = undefined;

      if (options != null) {
        if (isAbstractControlOptions(options)) {
          // `options` are `AbstractControlOptions`
          validators = options.validators != null ? options.validators : null;
          asyncValidators = options.asyncValidators != null ? options.asyncValidators : null;
          updateOn = options.updateOn != null ? options.updateOn : undefined;
        } else {
          // `options` are legacy form group options
          validators = options['validator'] != null ? options['validator'] : null;
          asyncValidators = options['asyncValidator'] != null ? options['asyncValidator'] : null;
        }
      }

      return new FormGroup(controls, {
        asyncValidators,
        updateOn,
        validators
      });
    }
    /**
     * @description
     * Construct a new `FormControl` with the given state, validators and options.
     *
     * @param formState Initializes the control with an initial state value, or
     * with an object that contains both a value and a disabled status.
     *
     * @param validatorOrOpts A synchronous validator function, or an array of
     * such functions, or an `AbstractControlOptions` object that contains
     * validation functions and a validation trigger.
     *
     * @param asyncValidator A single async validator or array of async validator
     * functions.
     *
     * @usageNotes
     *
     * ### Initialize a control as disabled
     *
     * The following example returns a control with an initial value in a disabled state.
     *
     * <code-example path="forms/ts/formBuilder/form_builder_example.ts" region="disabled-control">
     * </code-example>
     */


    control(formState, validatorOrOpts, asyncValidator) {
      return new FormControl(formState, validatorOrOpts, asyncValidator);
    }
    /**
     * Constructs a new `FormArray` from the given array of configurations,
     * validators and options.
     *
     * @param controlsConfig An array of child controls or control configs. Each
     * child control is given an index when it is registered.
     *
     * @param validatorOrOpts A synchronous validator function, or an array of
     * such functions, or an `AbstractControlOptions` object that contains
     * validation functions and a validation trigger.
     *
     * @param asyncValidator A single async validator or array of async validator
     * functions.
     */


    array(controlsConfig, validatorOrOpts, asyncValidator) {
      const controls = controlsConfig.map(c => this._createControl(c));
      return new FormArray(controls, validatorOrOpts, asyncValidator);
    }
    /** @internal */


    _reduceControls(controlsConfig) {
      const controls = {};
      Object.keys(controlsConfig).forEach(controlName => {
        controls[controlName] = this._createControl(controlsConfig[controlName]);
      });
      return controls;
    }
    /** @internal */


    _createControl(controlConfig) {
      if (controlConfig instanceof FormControl || controlConfig instanceof FormGroup || controlConfig instanceof FormArray) {
        return controlConfig;
      } else if (Array.isArray(controlConfig)) {
        const value = controlConfig[0];
        const validator = controlConfig.length > 1 ? controlConfig[1] : null;
        const asyncValidator = controlConfig.length > 2 ? controlConfig[2] : null;
        return this.control(value, validator, asyncValidator);
      } else {
        return this.control(controlConfig);
      }
    }

  }

  const useFormBuilder = () => {
    const formBuilderRef = React.useRef(new FormBuilder());
    return formBuilderRef.current;
  };

  const useFormGroup = formGroup => {
    const [, state] = React.useState({});
    const [, reRender] = React.useState(0);
    const formGroupRef = React.useRef(formGroup || new FormGroup({}));
    const setFormGroup$ = new rxjs.Subject();
    const destroy$ = new rxjs.Subject();
    React.useEffect(() => {
      setFormGroup$.pipe(rxjs.takeUntil(destroy$), rxjs.filter(formGroup => formGroup !== formGroupRef.current), rxjs.tap(formGroup => {
        setImmediate(() => {
          formGroupRef.current = formGroup;
          formGroupRef.current.dispatchStateFn = state;
        });
      }), rxjs.delay(1), rxjs.tap(() => {
        reRender(Math.random());
      })).subscribe();
    }, []);
    return {
      formGroup: formGroupRef.current,
      setFormGroup: setFormGroup$.next.bind(setFormGroup$)
    };
  };

  const useObservableSubscribe = props => {
    const [state, setState] = React.useState({});
    const destroy$Ref = React.useRef(props.destroy$ ? props.destroy$ : new rxjs.Subject());
    const obs$ = props.obs$;
    const destroy$ = destroy$Ref.current;
    React.useEffect(() => {
      obs$.pipe(rxjs.takeUntil(destroy$), rxjs.tap(value => {
        setState(produce.produce(draft => {
          draft.value = value;
        }));
      }, error => {
        setState(produce.produce(draft => {
          draft.error = error;
        }));
      }, () => {
        setState(produce.produce(draft => {
          draft.completed = true;
        }));
      })).subscribe();
      return () => {
        destroy$.next();
        destroy$.complete();
      };
    }, [obs$, destroy$]);
    return state;
  };

  function toNumber(value) {
    return typeof value === 'number' ? value : parseInt(value, 10);
  }

  class AbstractValidatorDirective {
    constructor() {
      this._validator = nullValidator;
      this._onChange = void 0;
      this.inputName = void 0;
    }

    handleChanges(changes) {
      if (this.inputName in changes) {
        const input = this.normalizeInput(changes[this.inputName].currentValue);
        this._validator = this.createValidator(input);

        if (this._onChange) {
          this._onChange();
        }
      }
    }

    validate(control) {
      return this._validator(control);
    }

    registerOnValidatorChange(fn) {
      this._onChange = fn;
    }

  }

  class MaxValidator extends AbstractValidatorDirective {
    constructor(...args) {
      super(...args);
      this.max = void 0;
      this.inputName = 'max';

      this.normalizeInput = input => parseFloat(input);

      this.createValidator = max => maxValidator(max);
    }

    ngOnChanges(changes) {
      this.handleChanges(changes);
    }

  }
  /**
   * @description
   * Provider which adds `MinValidator` to the `NG_VALIDATORS` multi-provider list.
   */

  /**
   * A directive which installs the {@link MinValidator} for any `formControlName`,
   * `formControl`, or control with `ngModel` that also has a `min` attribute.
   *
   * @see [Form Validation](guide/form-validation)
   *
   * @usageNotes
   *
   * ### Adding a min validator
   *
   * The following example shows how to add a min validator to an input attached to an
   * ngModel binding.
   *
   * ```html
   * <input type="number" ngModel min="4">
   * ```
   *
   * @ngModule ReactiveFormsModule
   * @ngModule FormsModule
   * @publicApi
   */

  class MinValidator extends AbstractValidatorDirective {
    constructor(...args) {
      super(...args);
      this.min = void 0;
      this.inputName = 'min';

      this.normalizeInput = input => parseFloat(input);

      this.createValidator = min => minValidator(min);
    }

    /**
     * Declare `ngOnChanges` lifecycle hook at the main directive level (vs keeping it in base class)
     * to avoid differences in handling inheritance of lifecycle hooks between Ivy and ViewEngine in
     * AOT mode. This could be refactored once ViewEngine is removed.
     * @nodoc
     */
    ngOnChanges(changes) {
      this.handleChanges(changes);
    }

  }
  /**
   * @description
   * An interface implemented by classes that perform asynchronous validation.
   *
   * @usageNotes
   *
   * ### Provide a custom async validator directive
   *
   * The following example implements the `AsyncValidator` interface to create an
   * async validator directive with a custom error key.
   *
   * ```typescript
   * import { of } from 'rxjs';
   *
   * @Directive({
   *   selector: '[customAsyncValidator]',
   *   providers: [{provide: NG_ASYNC_VALIDATORS, useExisting: CustomAsyncValidatorDirective, multi:
   * true}]
   * })
   * class CustomAsyncValidatorDirective implements AsyncValidator {
   *   validate(control: AbstractControl): Observable<ValidationErrors|null> {
   *     return of({'custom': true});
   *   }
   * }
   * ```
   *
   * @publicApi
   */

  /**
   * @description
   * Provider which adds `RequiredValidator` to the `NG_VALIDATORS` multi-provider list.
   */

  /**
   * @description
   * Provider which adds `CheckboxRequiredValidator` to the `NG_VALIDATORS` multi-provider list.
   */

  /**
   * @description
   * A directive that adds the `required` validator to any controls marked with the
   * `required` attribute. The directive is provided with the `NG_VALIDATORS` multi-provider list.
   *
   * @see [Form Validation](guide/form-validation)
   *
   * @usageNotes
   *
   * ### Adding a required validator using template-driven forms
   *
   * ```
   * <input name="fullName" ngModel required>
   * ```
   *
   * @ngModule FormsModule
   * @ngModule ReactiveFormsModule
   * @publicApi
   */
  class RequiredValidator {
    constructor() {
      this._required = false;
      this._onChange = void 0;
    }

    /**
     * @description
     * Tracks changes to the required attribute bound to this directive.
     */
    get required() {
      return this._required;
    }

    set required(value) {
      this._required = value != null && value !== false && `${value}` !== 'false';
      if (this._onChange) this._onChange();
    }
    /**
     * Method that validates whether the control is empty.
     * Returns the validation result if enabled, otherwise null.
     * @nodoc
     */


    validate(control) {
      return this.required ? requiredValidator(control) : null;
    }
    /**
     * Registers a callback function to call when the validator inputs change.
     * @nodoc
     */


    registerOnValidatorChange(fn) {
      this._onChange = fn;
    }

  }
  /**
   * A Directive that adds the `required` validator to checkbox controls marked with the
   * `required` attribute. The directive is provided with the `NG_VALIDATORS` multi-provider list.
   *
   * @see [Form Validation](guide/form-validation)
   *
   * @usageNotes
   *
   * ### Adding a required checkbox validator using template-driven forms
   *
   * The following example shows how to add a checkbox required validator to an input attached to an
   * ngModel binding.
   *
   * ```
   * <input type="checkbox" name="active" ngModel required>
   * ```
   *
   * @publicApi
   * @ngModule FormsModule
   * @ngModule ReactiveFormsModule
   */

  class CheckboxRequiredValidator extends RequiredValidator {
    /**
     * Method that validates whether or not the checkbox has been checked.
     * Returns the validation result if enabled, otherwise null.
     * @nodoc
     */
    //  override validate(control: AbstractControl): ValidationErrors|null {
    validate(control) {
      return this.required ? requiredTrueValidator(control) : null;
    }

  }
  /**
   * @description
   * Provider which adds `EmailValidator` to the `NG_VALIDATORS` multi-provider list.
   */

  /**
   * A directive that adds the `email` validator to controls marked with the
   * `email` attribute. The directive is provided with the `NG_VALIDATORS` multi-provider list.
   *
   * @see [Form Validation](guide/form-validation)
   *
   * @usageNotes
   *
   * ### Adding an email validator
   *
   * The following example shows how to add an email validator to an input attached to an ngModel
   * binding.
   *
   * ```
   * <input type="email" name="email" ngModel email>
   * <input type="email" name="email" ngModel email="true">
   * <input type="email" name="email" ngModel [email]="true">
   * ```
   *
   * @publicApi
   * @ngModule FormsModule
   * @ngModule ReactiveFormsModule
   */

  class EmailValidator {
    constructor() {
      this._enabled = false;
      this._onChange = void 0;
    }

    /**
     * @description
     * Tracks changes to the email attribute bound to this directive.
     */
    set email(value) {
      this._enabled = value === '' || value === true || value === 'true';
      if (this._onChange) this._onChange();
    }
    /**
     * Method that validates whether an email address is valid.
     * Returns the validation result if enabled, otherwise null.
     * @nodoc
     */


    validate(control) {
      return this._enabled ? emailValidator(control) : null;
    }
    /**
     * Registers a callback function to call when the validator inputs change.
     * @nodoc
     */


    registerOnValidatorChange(fn) {
      this._onChange = fn;
    }

  }
  /**
   * @description
   * A function that receives a control and synchronously returns a map of
   * validation errors if present, otherwise null.
   *
   * @publicApi
   */

  /**
   * @description
   * Provider which adds `MinLengthValidator` to the `NG_VALIDATORS` multi-provider list.
   */

  /**
   * A directive that adds minimum length validation to controls marked with the
   * `minlength` attribute. The directive is provided with the `NG_VALIDATORS` multi-provider list.
   *
   * @see [Form Validation](guide/form-validation)
   *
   * @usageNotes
   *
   * ### Adding a minimum length validator
   *
   * The following example shows how to add a minimum length validator to an input attached to an
   * ngModel binding.
   *
   * ```html
   * <input name="firstName" ngModel minlength="4">
   * ```
   *
   * @ngModule ReactiveFormsModule
   * @ngModule FormsModule
   * @publicApi
   */
  class MinLengthValidator {
    constructor() {
      this._validator = nullValidator;
      this._onChange = void 0;
      this.minlength = void 0;
    }

    // This input is always defined, since the name matches selector.

    /** @nodoc */
    ngOnChanges(changes) {
      if ('minlength' in changes) {
        this._createValidator();

        if (this._onChange) this._onChange();
      }
    }
    /**
     * Method that validates whether the value meets a minimum length requirement.
     * Returns the validation result if enabled, otherwise null.
     * @nodoc
     */


    validate(control) {
      return this.enabled() ? this._validator(control) : null;
    }
    /**
     * Registers a callback function to call when the validator inputs change.
     * @nodoc
     */


    registerOnValidatorChange(fn) {
      this._onChange = fn;
    }

    _createValidator() {
      this._validator = this.enabled() ? minLengthValidator(toNumber(this.minlength)) : nullValidator;
    }
    /** @nodoc */


    enabled() {
      return this.minlength != null
      /* both `null` and `undefined` */
      ;
    }

  }
  /**
   * @description
   * Provider which adds `MaxLengthValidator` to the `NG_VALIDATORS` multi-provider list.
   */

  /**
   * A directive that adds max length validation to controls marked with the
   * `maxlength` attribute. The directive is provided with the `NG_VALIDATORS` multi-provider list.
   *
   * @see [Form Validation](guide/form-validation)
   *
   * @usageNotes
   *
   * ### Adding a maximum length validator
   *
   * The following example shows how to add a maximum length validator to an input attached to an
   * ngModel binding.
   *
   * ```html
   * <input name="firstName" ngModel maxlength="25">
   * ```
   *
   * @ngModule ReactiveFormsModule
   * @ngModule FormsModule
   * @publicApi
   */

  class MaxLengthValidator {
    constructor() {
      this._validator = nullValidator;
      this._onChange = void 0;
      this.maxlength = void 0;
    }

    // This input is always defined, since the name matches selector.

    /** @nodoc */
    ngOnChanges(changes) {
      if ('maxlength' in changes) {
        this._createValidator();

        if (this._onChange) this._onChange();
      }
    }
    /**
     * Method that validates whether the value exceeds the maximum length requirement.
     * @nodoc
     */


    validate(control) {
      return this.enabled() ? this._validator(control) : null;
    }
    /**
     * Registers a callback function to call when the validator inputs change.
     * @nodoc
     */


    registerOnValidatorChange(fn) {
      this._onChange = fn;
    }

    _createValidator() {
      this._validator = this.enabled() ? maxLengthValidator(toNumber(this.maxlength)) : nullValidator;
    }
    /** @nodoc */


    enabled() {
      return this.maxlength != null
      /* both `null` and `undefined` */
      ;
    }

  }
  /**
   * @description
   * Provider which adds `PatternValidator` to the `NG_VALIDATORS` multi-provider list.
   */

  /**
   * @description
   * A directive that adds regex pattern validation to controls marked with the
   * `pattern` attribute. The regex must match the entire control value.
   * The directive is provided with the `NG_VALIDATORS` multi-provider list.
   *
   * @see [Form Validation](guide/form-validation)
   *
   * @usageNotes
   *
   * ### Adding a pattern validator
   *
   * The following example shows how to add a pattern validator to an input attached to an
   * ngModel binding.
   *
   * ```html
   * <input name="firstName" ngModel pattern="[a-zA-Z ]*">
   * ```
   *
   * @ngModule ReactiveFormsModule
   * @ngModule FormsModule
   * @publicApi
   */

  class PatternValidator {
    constructor() {
      this._validator = nullValidator;
      this._onChange = void 0;
      this.pattern = void 0;
    }

    // This input is always defined, since the name matches selector.

    /** @nodoc */
    ngOnChanges(changes) {
      if ('pattern' in changes) {
        this._createValidator();

        if (this._onChange) this._onChange();
      }
    }
    /**
     * Method that validates whether the value matches the pattern requirement.
     * @nodoc
     */


    validate(control) {
      return this._validator(control);
    }
    /**
     * Registers a callback function to call when the validator inputs change.
     * @nodoc
     */


    registerOnValidatorChange(fn) {
      this._onChange = fn;
    }

    _createValidator() {
      this._validator = patternValidator(this.pattern);
    }

  }

  exports.AbstractControl = AbstractControl;
  exports.AsyncCall = AsyncCall;
  exports.AsyncElement = AsyncElement;
  exports.CheckboxRequiredValidator = CheckboxRequiredValidator;
  exports.DISABLED = DISABLED;
  exports.EmailValidator = EmailValidator;
  exports.EventEmitter2 = EventEmitter2;
  exports.FormArray = FormArray;
  exports.FormControl = FormControl;
  exports.FormGroup = FormGroup;
  exports.INVALID = INVALID;
  exports.InputContainer = InputContainer;
  exports.MaxLengthValidator = MaxLengthValidator;
  exports.MaxValidator = MaxValidator;
  exports.MinLengthValidator = MinLengthValidator;
  exports.MinValidator = MinValidator;
  exports.MockServer = MockServer;
  exports.PENDING = PENDING;
  exports.PatternValidator = PatternValidator;
  exports.RequiredValidator = RequiredValidator;
  exports.VALID = VALID;
  exports.Validators = Validators;
  exports.addValidators = addValidators;
  exports.composeAsyncValidators = composeAsyncValidators;
  exports.composeValidators = composeValidators;
  exports.emailValidator = emailValidator;
  exports.getControlAsyncValidators = getControlAsyncValidators;
  exports.getControlValidators = getControlValidators;
  exports.getTypeOperation = getTypeOperation;
  exports.hasValidator = hasValidator;
  exports.initialStateAsyncCall = initialStateAsyncCall;
  exports.isEvent = isEvent;
  exports.isReactNative = isReactNative;
  exports.makeValidatorsArray = makeValidatorsArray;
  exports.maxLengthValidator = maxLengthValidator;
  exports.maxValidator = maxValidator;
  exports.mergeValidators = mergeValidators;
  exports.minLengthValidator = minLengthValidator;
  exports.minValidator = minValidator;
  exports.normalizeValidators = normalizeValidators;
  exports.nullValidator = nullValidator;
  exports.patternValidator = patternValidator;
  exports.removeValidators = removeValidators;
  exports.requiredTrueValidator = requiredTrueValidator;
  exports.requiredValidator = requiredValidator;
  exports.toObservable = toObservable;
  exports.useAsyncCall = useAsyncCall;
  exports.useDetroy$ = useDetroy$;
  exports.useEmitListen = useEmitListen;
  exports.useFormBuilder = useFormBuilder;
  exports.useFormGroup = useFormGroup;
  exports.useObservableSubscribe = useObservableSubscribe;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
