import { Subject, Subscription } from 'rxjs';

class EventEmitter_ extends Subject {
  // tslint:disable-line
  constructor(isAsync = false) {
    super();
    this.__isAsync = void 0;
    this.__isAsync = isAsync;
  }

  emit(value) {
    super.next(value);
  }

  subscribe(observerOrNext, error, complete) {
    let nextFn = observerOrNext;

    let errorFn = error || (() => null);

    let completeFn = complete;

    if (observerOrNext && typeof observerOrNext === 'object') {
      var _observer$next, _observer$error, _observer$complete;

      const observer = observerOrNext;
      nextFn = (_observer$next = observer.next) == null ? void 0 : _observer$next.bind(observer);
      errorFn = (_observer$error = observer.error) == null ? void 0 : _observer$error.bind(observer);
      completeFn = (_observer$complete = observer.complete) == null ? void 0 : _observer$complete.bind(observer);
    }

    if (this.__isAsync) {
      errorFn = _wrapInTimeout(errorFn);

      if (nextFn) {
        nextFn = _wrapInTimeout(nextFn);
      }

      if (completeFn) {
        completeFn = _wrapInTimeout(completeFn);
      }
    }

    const sink = super.subscribe({
      next: nextFn,
      error: errorFn,
      complete: completeFn
    });

    if (observerOrNext instanceof Subscription) {
      observerOrNext.add(sink);
    }

    return sink;
  }

}

function _wrapInTimeout(fn) {
  return value => {
    setTimeout(fn, undefined, value);
  };
}
/**
 * @publicApi
 */


export const EventEmitter2 = EventEmitter_;